#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.grib module."""

import os
import sys
from unittest import mock

import numpy as np
import pytest
import xarray as xr

from satpy.dataset import DataQuery

# Parameterized cases
TEST_ARGS = ("proj_params", "lon_corners", "lat_corners")
TEST_PARAMS = (
    (None, None, None),  # cyl default case
    (
        {
            "a": 6371229, "b": 6371229, "proj": "lcc",
            "lon_0": 265.0, "lat_0": 25.0,
            "lat_1": 25.0, "lat_2": 25.0
        },
        [-133.459, -65.12555139, -152.8786225, -49.41598659],
        [12.19, 14.34208538, 54.56534318, 57.32843565]
    ),
)


def fake_gribdata():
    """Return some faked data for use as grib values."""
    return np.arange(25.).reshape((5, 5))


def _round_trip_projection_lonlat_check(area):
    """Check that X/Y coordinates can be transformed multiple times.

    Many GRIB files include non-standard projects that work for the
    initial transformation of X/Y coordinates to longitude/latitude,
    but may fail in the reverse transformation. For example, an eqc
    projection that goes from 0 longitude to 360 longitude. The X/Y
    coordinates may accurately go from the original X/Y metered space
    to the correct longitude/latitude, but transforming those coordinates
    back to X/Y space will produce the wrong result.

    """
    from pyproj import Proj
    p = Proj(area.crs)
    x, y = area.get_proj_vectors()
    lon, lat = p(x, y, inverse=True)
    x2, y2 = p(lon, lat)
    np.testing.assert_almost_equal(x, x2)
    np.testing.assert_almost_equal(y, y2)


class FakeMessage(object):
    """Fake message returned by pygrib.open().message(x)."""

    def __init__(self, values, proj_params=None, latlons=None, **attrs):
        """Init the message."""
        super(FakeMessage, self).__init__()
        self.attrs = attrs
        self.values = values
        if proj_params is None:
            proj_params = {"a": 6371229, "b": 6371229, "proj": "cyl"}
        self.projparams = proj_params
        self._latlons = latlons

    def keys(self):
        """Get message keys."""
        return self.attrs.keys()

    def latlons(self):
        """Get coordinates."""
        return self._latlons

    def __getitem__(self, item):
        """Get item."""
        return self.attrs[item]

    def valid_key(self, key):
        """Validate key."""
        return True


class FakeGRIB(object):
    """Fake GRIB file returned by pygrib.open."""

    def __init__(self, messages=None, proj_params=None, latlons=None):
        """Init the grib file."""
        super(FakeGRIB, self).__init__()
        if messages is not None:
            self._messages = messages
        else:
            self._messages = [
                FakeMessage(
                    values=fake_gribdata(),
                    name="TEST",
                    shortName="t",
                    level=100,
                    pressureUnits="hPa",
                    cfName="air_temperature",
                    units="K",
                    dataDate=20180504,
                    dataTime=1200,
                    validityDate=20180504,
                    validityTime=1800,
                    distinctLongitudes=np.arange(5.),
                    distinctLatitudes=np.arange(5.),
                    missingValue=9999,
                    modelName="notknown",
                    minimum=100.,
                    maximum=200.,
                    typeOfLevel="isobaricInhPa",
                    jScansPositively=0,
                    proj_params=proj_params,
                    latlons=latlons,
                ),
                FakeMessage(
                    values=fake_gribdata(),
                    name="TEST",
                    shortName="t",
                    level=200,
                    pressureUnits="hPa",
                    cfName="air_temperature",
                    units="K",
                    dataDate=20180504,
                    dataTime=1200,
                    validityDate=20180504,
                    validityTime=1800,
                    distinctLongitudes=np.arange(5.),
                    distinctLatitudes=np.arange(5.),
                    missingValue=9999,
                    modelName="notknown",
                    minimum=100.,
                    maximum=200.,
                    typeOfLevel="isobaricInhPa",
                    jScansPositively=1,
                    proj_params=proj_params,
                    latlons=latlons,
                ),
                FakeMessage(
                    values=fake_gribdata(),
                    name="TEST",
                    shortName="t",
                    level=300,
                    pressureUnits="hPa",
                    cfName="air_temperature",
                    units="K",
                    dataDate=20180504,
                    dataTime=1200,
                    validityDate=20180504,
                    validityTime=1800,
                    distinctLongitudes=np.arange(5.),
                    distinctLatitudes=np.arange(5.),
                    missingValue=9999,
                    minimum=100.,
                    maximum=200.,
                    typeOfLevel="isobaricInhPa",
                    jScansPositively=0,
                    proj_params=proj_params,
                    latlons=latlons,
                ),
            ]
        self.messages = len(self._messages)

    def message(self, msg_num):
        """Get a message."""
        return self._messages[msg_num - 1]

    def seek(self, loc):
        """Seek."""
        return

    def __iter__(self):
        """Iterate."""
        return iter(self._messages)

    def __enter__(self):
        """Enter."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit."""


class TestGRIBReader:
    """Test GRIB Reader."""

    yaml_file = "grib.yaml"

    def setup_method(self):
        """Wrap pygrib to read fake data."""
        from satpy._config import config_search_paths
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))

        try:
            import pygrib
        except ImportError:
            pygrib = None
        self.orig_pygrib = pygrib
        sys.modules["pygrib"] = mock.MagicMock()

    def teardown_method(self):
        """Re-enable pygrib import."""
        sys.modules["pygrib"] = self.orig_pygrib

    def _get_test_datasets(self, dataids, fake_pygrib=None):
        from satpy.readers.core.loading import load_reader
        if fake_pygrib is None:
            fake_pygrib = FakeGRIB()

        with mock.patch("satpy.readers.grib.pygrib") as pg:
            pg.open.return_value = fake_pygrib
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames([
                "gfs.t18z.sfluxgrbf106.grib2",
            ])
            r.create_filehandlers(loadables)
            datasets = r.load(dataids)
        return datasets

    @staticmethod
    def _get_fake_pygrib(proj_params, lon_corners, lat_corners):
        latlons = None
        if lon_corners is not None:
            lats = np.array([
                [lat_corners[0], 0, 0, 0, lat_corners[1]],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [lat_corners[2], 0, 0, 0, lat_corners[3]]])
            lons = np.array([
                [lon_corners[0], 0, 0, 0, lon_corners[1]],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [lon_corners[2], 0, 0, 0, lon_corners[3]]])
            latlons = (lats, lons)

        fake_pygrib = FakeGRIB(
            proj_params=proj_params,
            latlons=latlons)
        return fake_pygrib

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        with mock.patch("satpy.readers.grib.pygrib") as pg:
            pg.open.return_value = FakeGRIB()
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames([
                "gfs.t18z.sfluxgrbf106.grib2",
            ])
            assert len(loadables) == 1
            r.create_filehandlers(loadables)
            # make sure we have some files
            assert r.file_handlers

    def test_file_pattern(self):
        """Test matching of file patterns."""
        from satpy.readers.core.loading import load_reader

        filenames = [
                "quinoa.grb",
                "tempeh.grb2",
                "tofu.grib2",
                "falafel.grib",
                "S_NWC_NWP_1900-01-01T00:00:00Z_999.grib"]

        r = load_reader(self.reader_configs)
        files = r.select_files_from_pathnames(filenames)
        assert len(files) == 4

    @pytest.mark.parametrize(TEST_ARGS, TEST_PARAMS)
    def test_load_all(self, proj_params, lon_corners, lat_corners):
        """Test loading all test datasets."""
        fake_pygrib = self._get_fake_pygrib(proj_params, lon_corners, lat_corners)
        dataids = [
            DataQuery(name="t", level=100, modifiers=tuple()),
            DataQuery(name="t", level=200, modifiers=tuple()),
            DataQuery(name="t", level=300, modifiers=tuple())
        ]
        datasets = self._get_test_datasets(dataids, fake_pygrib)

        assert len(datasets) == 3
        for v in datasets.values():
            assert v.attrs["units"] == "K"
            assert isinstance(v, xr.DataArray)

    @pytest.mark.parametrize(TEST_ARGS, TEST_PARAMS)
    def test_area_def_crs(self, proj_params, lon_corners, lat_corners):
        """Check that the projection is accurate."""
        fake_pygrib = self._get_fake_pygrib(proj_params, lon_corners, lat_corners)
        dataids = [DataQuery(name="t", level=100, modifiers=tuple())]
        datasets = self._get_test_datasets(dataids, fake_pygrib)
        area = datasets["t"].attrs["area"]
        if not hasattr(area, "crs"):
            pytest.skip("Can't test with pyproj < 2.0")
        _round_trip_projection_lonlat_check(area)

    @pytest.mark.parametrize(TEST_ARGS, TEST_PARAMS)
    def test_missing_attributes(self, proj_params, lon_corners, lat_corners):
        """Check that the grib reader handles missing attributes in the grib file."""
        fake_pygrib = self._get_fake_pygrib(proj_params, lon_corners, lat_corners)

        # This has modelName
        query_contains = DataQuery(name="t", level=100, modifiers=tuple())
        # This does not have modelName
        query_not_contains = DataQuery(name="t", level=300, modifiers=tuple())
        dataset = self._get_test_datasets([query_contains, query_not_contains], fake_pygrib)
        assert dataset[query_contains].attrs["modelName"] == "notknown"
        assert dataset[query_not_contains].attrs["modelName"] == "unknown"

    @pytest.mark.parametrize(TEST_ARGS, TEST_PARAMS)
    def test_jscanspositively(self, proj_params, lon_corners, lat_corners):
        """Check that data is flipped if the jScansPositively is present."""
        fake_pygrib = self._get_fake_pygrib(proj_params, lon_corners, lat_corners)

        # This has no jScansPositively
        query_not_contains = DataQuery(name="t", level=100, modifiers=tuple())
        # This contains jScansPositively
        query_contains = DataQuery(name="t", level=200, modifiers=tuple())
        dataset = self._get_test_datasets([query_contains, query_not_contains], fake_pygrib)

        np.testing.assert_allclose(fake_gribdata(), dataset[query_not_contains].values)
        np.testing.assert_allclose(fake_gribdata(), dataset[query_contains].values[::-1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test the MSG common (native and hrit format) functionionalities."""

import datetime as dt
import unittest

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.readers.core.seviri import (
    MEIRINK_COEFS,
    MEIRINK_EPOCH,
    MeirinkCoefficients,
    NoValidOrbitParams,
    OrbitPolynomial,
    OrbitPolynomialFinder,
    chebyshev,
    dec10216,
    get_cds_time,
    get_padding_area,
    get_satpos,
    pad_data_horizontally,
    pad_data_vertically,
    round_nom_time,
)
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()


def chebyshev4(c, x, domain):
    """Evaluate 4th order Chebyshev polynomial."""
    start_x, end_x = domain
    t = (x - 0.5 * (end_x + start_x)) / (0.5 * (end_x - start_x))
    return c[0] + c[1]*t + c[2]*(2*t**2 - 1) + c[3]*(4*t**3 - 3*t) - 0.5*c[0]


class SeviriBaseTest(unittest.TestCase):
    """Test SEVIRI base."""

    def test_dec10216(self):
        """Test the dec10216 function."""
        res = dec10216(np.array([255, 255, 255, 255, 255], dtype=np.uint8))
        exp = (np.ones((4, )) * 1023).astype(np.uint16)
        np.testing.assert_equal(res, exp)
        res = dec10216(np.array([1, 1, 1, 1, 1], dtype=np.uint8))
        exp = np.array([4,  16,  64, 257], dtype=np.uint16)
        np.testing.assert_equal(res, exp)

    def test_chebyshev(self):
        """Test the chebyshev function."""
        coefs = [1, 2, 3, 4]
        time = 123
        domain = [120, 130]
        res = chebyshev(coefs=[1, 2, 3, 4], time=time, domain=domain)
        exp = chebyshev4(coefs, time, domain)
        np.testing.assert_allclose(res, exp)

    def test_get_cds_time_scalar(self):
        """Test the get_cds_time function for scalar inputs."""
        assert get_cds_time(days=21246, msecs=12 * 3600 * 1000) == np.datetime64("2016-03-03 12:00")

    def test_get_cds_time_array(self):
        """Test the get_cds_time function for array inputs."""
        days = np.array([21246, 21247, 21248])
        msecs = np.array([12*3600*1000, 13*3600*1000 + 1, 14*3600*1000 + 2])
        expected = np.array([np.datetime64("2016-03-03 12:00:00.000"),
                             np.datetime64("2016-03-04 13:00:00.001"),
                             np.datetime64("2016-03-05 14:00:00.002")])
        res = get_cds_time(days=days, msecs=msecs)
        np.testing.assert_equal(res, expected)

    def test_get_cds_time_nanoseconds(self):
        """Test the get_cds_time function for having nanosecond precision."""
        days = 21246
        msecs = 12 * 3600 * 1000
        expected = np.datetime64("2016-03-03 12:00:00.000")
        res = get_cds_time(days=days, msecs=msecs)
        np.testing.assert_equal(res, expected)
        assert res.dtype == np.dtype("datetime64[ns]")

    def test_pad_data_horizontally_bad_shape(self):
        """Test the error handling for the horizontal hrv padding."""
        data = xr.DataArray(data=np.zeros((1, 10)), dims=("y", "x"))
        east_bound = 5
        west_bound = 10
        final_size = (1, 20)
        with pytest.raises(IndexError):
            pad_data_horizontally(data, final_size, east_bound, west_bound)

    def test_pad_data_vertically_bad_shape(self):
        """Test the error handling for the vertical hrv padding."""
        data = xr.DataArray(data=np.zeros((10, 1)), dims=("y", "x"))
        south_bound = 5
        north_bound = 10
        final_size = (20, 1)
        with pytest.raises(IndexError):
            pad_data_vertically(data, final_size, south_bound, north_bound)

    def observation_start_time(self):
        """Get scan start timestamp for testing."""
        return dt.datetime(2023, 3, 20, 15, 0, 10, 691000)

    def observation_end_time(self):
        """Get scan end timestamp for testing."""
        return dt.datetime(2023, 3, 20, 15, 12, 43, 843000)

    def test_round_nom_time(self):
        """Test the rouding of start/end_time."""
        assert round_nom_time(date=self.observation_start_time(),
                              time_delta=dt.timedelta(minutes=15)) == dt.datetime(2023, 3, 20, 15, 0)
        assert round_nom_time(date=self.observation_end_time(),
                              time_delta=dt.timedelta(minutes=15)) == dt.datetime(2023, 3, 20, 15, 15)

    @staticmethod
    def test_pad_data_horizontally():
        """Test the horizontal hrv padding."""
        data = xr.DataArray(data=np.zeros((1, 10)), dims=("y", "x"))
        east_bound = 4
        west_bound = 13
        final_size = (1, 20)
        res = pad_data_horizontally(data, final_size, east_bound, west_bound)
        expected = np.array([[np.nan, np.nan, np.nan,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                              np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]])
        np.testing.assert_equal(res, expected)

    @staticmethod
    def test_pad_data_vertically():
        """Test the vertical hrv padding."""
        data = xr.DataArray(data=np.zeros((10, 1)), dims=("y", "x"))
        south_bound = 4
        north_bound = 13
        final_size = (20, 1)
        res = pad_data_vertically(data, final_size, south_bound, north_bound)
        expected = np.zeros(final_size)
        expected[:] = np.nan
        expected[south_bound-1:north_bound] = 0.
        np.testing.assert_equal(res, expected)

    @staticmethod
    def test_get_padding_area_float():
        """Test padding area generator for floats."""
        shape = (10, 10)
        dtype = np.float64
        res = get_padding_area(shape, dtype)
        expected = da.full(shape, np.nan, dtype=dtype, chunks=CHUNK_SIZE)
        np.testing.assert_array_equal(res, expected)

    @staticmethod
    def test_get_padding_area_int():
        """Test padding area generator for integers."""
        shape = (10, 10)
        dtype = np.int64
        res = get_padding_area(shape, dtype)
        expected = da.full(shape, 0, dtype=dtype, chunks=CHUNK_SIZE)
        np.testing.assert_array_equal(res, expected)


ORBIT_POLYNOMIALS = {
    "StartTime": np.array([
        [
            dt.datetime(2006, 1, 1, 6), dt.datetime(2006, 1, 1, 12),
            dt.datetime(2006, 1, 1, 18), dt.datetime(1958, 1, 1, 0)]
    ]),
    "EndTime": np.array([
        [
            dt.datetime(2006, 1, 1, 12), dt.datetime(2006, 1, 1, 18),
            dt.datetime(2006, 1, 2, 0), dt.datetime(1958, 1, 1, 0)
        ]
    ]),
    "X": [np.zeros(8),
          [8.41607082e+04, 2.94319260e+00, 9.86748617e-01,
           -2.70135453e-01,
           -3.84364650e-02, 8.48718433e-03, 7.70548174e-04,
           -1.44262718e-04],
          np.zeros(8)],
    "Y": [np.zeros(8),
          [-5.21170255e+03, 5.12998948e+00, -1.33370453e+00,
           -3.09634144e-01,
           6.18232793e-02, 7.50505681e-03, -1.35131011e-03,
           -1.12054405e-04],
          np.zeros(8)],
    "Z": [np.zeros(8),
          [-6.51293855e+02, 1.45830459e+02, 5.61379400e+01,
           -3.90970565e+00,
           -7.38137565e-01, 3.06131644e-02, 3.82892428e-03,
           -1.12739309e-04],
          np.zeros(8)],
}
ORBIT_POLYNOMIALS_SYNTH = {
    # 12-31: Contiguous
    # 01-01: Small gap (12:00 - 13:00)
    # 01-02: Large gap (04:00 - 18:00)
    # 01-03: Overlap (10:00 - 13:00)
    "StartTime": np.array([
        [
            dt.datetime(2005, 12, 31, 10), dt.datetime(2005, 12, 31, 12),
            dt.datetime(2006, 1, 1, 10), dt.datetime(2006, 1, 1, 13),
            dt.datetime(2006, 1, 2, 0), dt.datetime(2006, 1, 2, 18),
            dt.datetime(2006, 1, 3, 6), dt.datetime(2006, 1, 3, 10),
        ]
    ]),
    "EndTime": np.array([
        [
            dt.datetime(2005, 12, 31, 12), dt.datetime(2005, 12, 31, 18),
            dt.datetime(2006, 1, 1, 12), dt.datetime(2006, 1, 1, 18),
            dt.datetime(2006, 1, 2, 4), dt.datetime(2006, 1, 2, 22),
            dt.datetime(2006, 1, 3, 13), dt.datetime(2006, 1, 3, 18),
        ]
    ]),
    "X": [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0],
    "Y": [1.1, 2.1, 3.1, 4.1, 5.1, 6.1, 7.1, 8.1],
    "Z": [1.2, 2.2, 3.2, 4.2, 5.2, 6.2, 7.2, 8.2],
}
ORBIT_POLYNOMIALS_INVALID = {
    "StartTime": np.array([
        [
            dt.datetime(1958, 1, 1), dt.datetime(1958, 1, 1)
        ]
    ]),
    "EndTime": np.array([
        [
            dt.datetime(1958, 1, 1), dt.datetime(1958, 1, 1)
        ]
    ]),
    "X": [1, 2],
    "Y": [3, 4],
    "Z": [5, 6],
}


class TestSatellitePosition:
    """Test locating the satellite."""

    @pytest.fixture
    def orbit_polynomial(self):
        """Get an orbit polynomial for testing."""
        return OrbitPolynomial(
            start_time=dt.datetime(2006, 1, 1, 12),
            end_time=dt.datetime(2006, 1, 1, 18),
            coefs=(
                np.array([8.41607082e+04, 2.94319260e+00, 9.86748617e-01,
                          -2.70135453e-01, -3.84364650e-02, 8.48718433e-03,
                          7.70548174e-04, -1.44262718e-04]),
                np.array([-5.21170255e+03, 5.12998948e+00, -1.33370453e+00,
                          -3.09634144e-01, 6.18232793e-02, 7.50505681e-03,
                          -1.35131011e-03, -1.12054405e-04]),
                np.array([-6.51293855e+02, 1.45830459e+02, 5.61379400e+01,
                          -3.90970565e+00, -7.38137565e-01, 3.06131644e-02,
                          3.82892428e-03, -1.12739309e-04])
            )
        )

    @pytest.fixture
    def time(self):
        """Get scan timestamp for testing."""
        return dt.datetime(2006, 1, 1, 12, 15, 9, 304888)

    def test_eval_polynomial(self, orbit_polynomial, time):
        """Test getting the position in cartesian coordinates."""
        x, y, z = orbit_polynomial.evaluate(time)
        np.testing.assert_allclose(
            [x, y, z],
            [42078421.37095518, -2611352.744615312, -419828.9699940758]
        )

    def test_get_satpos(self, orbit_polynomial, time):
        """Test getting the position in geodetic coordinates."""
        lon, lat, alt = get_satpos(
            orbit_polynomial=orbit_polynomial,
            time=time,
            semi_major_axis=6378169.00,
            semi_minor_axis=6356583.80
        )
        np.testing.assert_allclose(
            [lon, lat, alt],
            [-3.55117540817073, -0.5711243456528018, 35783296.150123544]
        )


class TestOrbitPolynomialFinder:
    """Unit tests for orbit polynomial finder."""

    @pytest.mark.parametrize(
        ("orbit_polynomials", "time", "orbit_polynomial_exp"),
        [
            # Contiguous validity intervals (that's the norm)
            (
                ORBIT_POLYNOMIALS_SYNTH,
                dt.datetime(2005, 12, 31, 12, 15),
                OrbitPolynomial(
                    coefs=(2.0, 2.1, 2.2),
                    start_time=np.datetime64("2005-12-31 12:00"),
                    end_time=np.datetime64("2005-12-31 18:00")
                )
            ),
            # No interval enclosing the given timestamp, but closest interval
            # not too far away
            (
                    ORBIT_POLYNOMIALS_SYNTH,
                    dt.datetime(2006, 1, 1, 12, 15),
                    OrbitPolynomial(
                        coefs=(3.0, 3.1, 3.2),
                        start_time=np.datetime64("2006-01-01 10:00"),
                        end_time=np.datetime64("2006-01-01 12:00")
                    )
            ),
            # Overlapping intervals
            (
                    ORBIT_POLYNOMIALS_SYNTH,
                    dt.datetime(2006, 1, 3, 12, 15),
                    OrbitPolynomial(
                        coefs=(8.0, 8.1, 8.2),
                        start_time=np.datetime64("2006-01-03 10:00"),
                        end_time=np.datetime64("2006-01-03 18:00")
                    )
            ),
        ]
    )
    def test_get_orbit_polynomial(self, orbit_polynomials, time,
                                  orbit_polynomial_exp):
        """Test getting the satellite locator."""
        import warnings
        finder = OrbitPolynomialFinder(orbit_polynomials)
        with warnings.catch_warnings():
            # There's no exact polynomial time match, filter the warning
            warnings.filterwarnings("ignore", category=UserWarning, message=r"No orbit polynomial valid")
            orbit_polynomial = finder.get_orbit_polynomial(time=time)
        assert orbit_polynomial == orbit_polynomial_exp

    @pytest.mark.parametrize(
        ("orbit_polynomials", "time"),
        [
            # No interval enclosing the given timestamp and closest interval
            # too far away
            (ORBIT_POLYNOMIALS_SYNTH, dt.datetime(2006, 1, 2, 12, 15)),
            # No valid polynomials at all
            (ORBIT_POLYNOMIALS_INVALID, dt.datetime(2006, 1, 1, 12, 15))
        ]
    )
    def test_get_orbit_polynomial_exceptions(self, orbit_polynomials, time):
        """Test exceptions thrown while getting the satellite locator."""
        finder = OrbitPolynomialFinder(orbit_polynomials)
        with pytest.raises(NoValidOrbitParams):
            with pytest.warns(UserWarning, match=r"No orbit polynomial valid"):
                finder.get_orbit_polynomial(time=time)


class TestMeirinkSlope:
    """Unit tests for the slope of Meirink calibration."""

    @pytest.mark.parametrize("platform_id", [321, 322, 323, 324])
    @pytest.mark.parametrize("channel_name", ["VIS006", "VIS008", "IR_016"])
    def test_get_meirink_slope_epoch(self, platform_id, channel_name):
        """Test the value of the slope of the Meirink calibration on 2000-01-01."""
        comp = MeirinkCoefficients(platform_id, channel_name, MEIRINK_EPOCH)
        coefs = comp.get_coefs("dummy_offset")
        assert coefs["MEIRINK-2023"][channel_name]["gain"] == MEIRINK_COEFS["2023"][platform_id][channel_name][0]/1000.

    @pytest.mark.parametrize(("platform_id", "time", "expected"), [
        (321, dt.datetime(2005, 1, 18, 0, 0), [0.0250354716, 0.0315626684, 0.022880986]),
        (321, dt.datetime(2010, 12, 31, 0, 0), [0.0258479563, 0.0322386887, 0.022895110500000003]),
        (322, dt.datetime(2010, 1, 18, 0, 0), [0.021964051999999998, 0.027548445, 0.021576766]),
        (322, dt.datetime(2015, 6, 1, 0, 0), [0.022465028, 0.027908105, 0.021674373999999996]),
        (323, dt.datetime(2005, 1, 18, 0, 0), [0.0209088464, 0.0265355228, 0.0230132616]),
        (323, dt.datetime(2010, 12, 31, 0, 0), [0.022181355200000002, 0.0280103379, 0.0229511138]),
        (324, dt.datetime(2010, 1, 18, 0, 0), [0.0218362, 0.027580748, 0.022285370999999998]),
        (324, dt.datetime(2015, 6, 1, 0, 0), [0.0225418, 0.028530172, 0.022248718999999997]),
    ])
    def test_get_meirink_slope_2020(self, platform_id, time, expected):
        """Test the value of the slope of the Meirink calibration."""
        for i, channel_name in enumerate(["VIS006", "VIS008", "IR_016"]):
            comp = MeirinkCoefficients(platform_id, channel_name, time)
            coefs = comp.get_coefs("dummy_offset")
            assert abs(coefs["MEIRINK-2023"][channel_name]["gain"] - expected[i]) < 1e-6



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""AHI specific enhancements."""

from __future__ import annotations

import logging

import dask.array as da
import numpy as np

from .wrappers import exclude_alpha, on_dask_array

LOG = logging.getLogger(__name__)


def jma_true_color_reproduction(img):
    """Apply CIE XYZ matrix and return True Color Reproduction data.

    Himawari-8 True Color Reproduction Approach Based on the CIE XYZ Color System
    Hidehiko MURATA, Kotaro SAITOH, and Yasuhiko SUMIDA
    Meteorological Satellite Center, Japan Meteorological Agency
    NOAA National Environmental Satellite, Data, and Information Service
    Colorado State University—CIRA
    https://www.jma.go.jp/jma/jma-eng/satellite/introduction/TCR.html
    """
    _jma_true_color_reproduction(img.data,
                                 platform=img.data.attrs["platform_name"])


@exclude_alpha
@on_dask_array
def _jma_true_color_reproduction(img_data, platform=None):
    """Convert from AHI RGB space to sRGB space.

    The conversion matrices for this are supplied per-platform.
    The matrices are computed using the method described in the paper:
    'True Color Imagery Rendering for Himawari-8 with a Color Reproduction Approach
    Based on the CIE XYZ Color System' (:doi:`10.2151/jmsj.2018-049`).

    """
    # Conversion matrix dictionaries specifying sensor and platform.
    ccm_dict = {"himawari-8": np.array([[1.1629, 0.1539, -0.2175],
                                        [-0.0252, 0.8725, 0.1300],
                                        [-0.0204, -0.1100, 1.0633]]),

                "himawari-9": np.array([[1.1619, 0.1542, -0.2168],
                                        [-0.0271, 0.8749, 0.1295],
                                        [-0.0202, -0.1103, 1.0634]]),

                "goes-16": np.array([[1.1425, 0.1819, -0.2250],
                                     [-0.0951, 0.9363, 0.1360],
                                     [-0.0113, -0.1179, 1.0621]]),
                "goes-17": np.array([[1.1437, 0.1818, -0.2262],
                                     [-0.0952, 0.9354, 0.1371],
                                     [-0.0113, -0.1178, 1.0620]]),
                "goes-18": np.array([[1.1629, 0.1539, -0.2175],
                                     [-0.0252, 0.8725, 0.1300],
                                     [-0.0204, -0.1100, 1.0633]]),
                "goes-19": np.array([[0.9481, 0.3706, -0.2194],
                                     [-0.0150, 0.8605, 0.1317],
                                     [-0.0174, -0.1009, 1.0512]]),

                "mtg-i1": np.array([[0.9007, 0.2086, -0.0100],
                                    [-0.0475, 1.0662, -0.0414],
                                    [-0.0123, -0.1342, 1.0794]]),

                "geo-kompsat-2a": np.array([[1.1661, 0.1489, -0.2157],
                                            [-0.0255, 0.8745, 0.1282],
                                            [-0.0205, -0.1103, 1.0637]]),
                }

    # A conversion matrix, sensor name and platform name is required
    if platform is None:
        raise ValueError("Missing platform name.")

    # Get the satellite-specific conversion matrix
    try:
        ccm = ccm_dict[platform.lower()]
    except KeyError:
        raise KeyError(f"No conversion matrix found for platform {platform}")

    output = da.dot(img_data.T, ccm.T)
    return output.T



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""HRIT/LRIT format reader.

This module is the base module for all HRIT-based formats. Here, you will find
the common building blocks for hrit reading.

One of the features here is the on-the-fly decompression of hrit files when
compressed hrit files are encountered (files finishing with `.C_`).
"""
from __future__ import annotations

import datetime as dt
import logging
import typing
from typing import Any

import dask
import dask.array as da
import numpy as np
import xarray as xr
from pyresample import geometry

import satpy.readers.core.utils as utils
from satpy.readers.core.eum import time_cds_short
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.seviri import dec10216

if typing.TYPE_CHECKING:
    import io

logger = logging.getLogger("hrit_base")

common_hdr = np.dtype([("hdr_id", "u1"),
                       ("record_length", ">u2")])

primary_header = np.dtype([("file_type", "u1"),
                           ("total_header_length", ">u4"),
                           ("data_field_length", ">u8")])

image_structure = np.dtype([("number_of_bits_per_pixel", "u1"),
                            ("number_of_columns", ">u2"),
                            ("number_of_lines", ">u2"),
                            ("compression_flag_for_data", "u1")])

image_navigation = np.dtype([("projection_name", "S32"),
                             ("cfac", ">i4"),
                             ("lfac", ">i4"),
                             ("coff", ">i4"),
                             ("loff", ">i4")])

image_data_function = np.dtype([("function", "|S1")])

annotation_header = np.dtype([("annotation", "|S1")])

timestamp_record = np.dtype([("cds_p_field", "u1"),
                             ("timestamp", time_cds_short)])

ancillary_text = np.dtype([("ancillary", "|S1")])

key_header = np.dtype([("key", "|S1")])

base_text_headers = {image_data_function: "image_data_function",
                     annotation_header: "annotation_header",
                     ancillary_text: "ancillary_text",
                     key_header: "key_header"}

base_hdr_map = {0: primary_header,
                1: image_structure,
                2: image_navigation,
                3: image_data_function,
                4: annotation_header,
                5: timestamp_record,
                6: ancillary_text,
                7: key_header,
                }


def decompress_file(infile) -> bytes:
    """Decompress an XRIT data file and return the decompressed buffer."""
    # decompress in-memory
    with open(infile, mode="rb") as fh:
        return decompress_buffer(fh.read())


def decompress_buffer(buffer) -> bytes:
    """Decompress buffer."""
    from pyPublicDecompWT import xRITDecompress
    xrit = xRITDecompress()
    xrit.decompress(buffer)

    return xrit.data()


def get_header_id(fp: io.BufferedReader) -> np.ndarray:
    """Return the HRIT header common data."""
    data = fp.read(common_hdr.itemsize)
    return np.frombuffer(data, dtype=common_hdr, count=1)[0]


def get_header_content(fp, header_dtype, count=1):
    """Return the content of the HRIT header."""
    data = fp.read(header_dtype.itemsize * count)
    return np.frombuffer(data, dtype=header_dtype, count=count)


def parse_header_dict(
        fp: io.BufferedReader,
        hdr_info: tuple[dict[int, np.dtype], dict[np.dtype, str], dict[np.dtype, str]],
        verbose: bool = False,
) -> dict[str, Any]:
    """Parse HRIT header data into a dictionary."""
    hdr_map, variable_length_headers, text_headers = hdr_info

    mda: dict[str, Any] = {}
    total_header_length = 16
    while fp.tell() < total_header_length:
        hdr_id = get_header_id(fp)
        if verbose:
            print("hdr_id")  # noqa: T201
            print(f'np.void({hdr_id}, dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),')  # noqa: T201
        the_type = hdr_map[hdr_id["hdr_id"]]
        if the_type in variable_length_headers:
            field_length = int((hdr_id["record_length"] - 3) /
                               the_type.itemsize)
            current_hdr = get_header_content(fp, the_type, field_length)
            if verbose:
                print(f"np.zeros(({field_length}, ), dtype={the_type}),")  # noqa: T201
            key = variable_length_headers[the_type]
            if key in mda:
                if not isinstance(mda[key], list):
                    mda[key] = [mda[key]]
                mda[key].append(current_hdr)
            else:
                mda[key] = current_hdr
        elif the_type in text_headers:
            field_length = int((hdr_id["record_length"] - 3) /
                               the_type.itemsize)
            char = list(the_type.fields.values())[0][0].char
            new_type = np.dtype(char + str(field_length))
            current_hdr = get_header_content(fp, new_type)[0]
            if verbose:
                print(f'np.array({current_hdr}, dtype="{new_type}"),')  # noqa: T201
            mda[text_headers[the_type]] = current_hdr
        else:
            current_hdr = get_header_content(fp, the_type)[0]
            if verbose:
                print(f"np.void({current_hdr}, dtype={the_type}),")  # noqa: T201
            mda.update(dict(zip(current_hdr.dtype.names, current_hdr)))

        total_header_length = mda["total_header_length"]
    return mda


class HRITFileHandler(BaseFileHandler):
    """HRIT standard format reader."""

    def __init__(self, filename, filename_info, filetype_info, hdr_info):
        """Initialize the reader."""
        super().__init__(filename, filename_info, filetype_info)

        self.mda = {}
        self.hdr_info = hdr_info
        self.mda = self._get_hd(self.hdr_info)
        self._start_time = filename_info["start_time"]
        self._end_time = self._start_time + dt.timedelta(minutes=15)

    def _get_hd(self, hdr_info, verbose=False):
        """Open the file, read and get the basic file header info and set the mda dictionary."""
        with utils.generic_open(self.filename, mode="rb") as fp:
            mda = parse_header_dict(fp, hdr_info, verbose=verbose)

        mda.setdefault("number_of_bits_per_pixel", 10)

        mda["projection_parameters"] = {"a": 6378169.00,
                                             "b": 6356583.80,
                                             "h": 35785831.00,
                                             # FIXME: find a reasonable SSP
                                             "SSP_longitude": 0.0}
        mda["orbital_parameters"] = {}
        return mda

    @property
    def observation_start_time(self):
        """Get observation start time."""
        return self._start_time

    @property
    def observation_end_time(self):
        """Get observation end time."""
        return self._end_time

    @property
    def start_time(self):
        """Get start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get end time."""
        return self._end_time

    def get_dataset(self, key, info):
        """Load a dataset."""
        # Read bands
        data = self.read_band(key, info)

        # Convert to xarray
        xdata = xr.DataArray(data, dims=["y", "x"])

        return xdata

    def get_xy_from_linecol(self, line, col, offsets, factors):
        """Get the intermediate coordinates from line & col.

        Intermediate coordinates are actually the instruments scanning angles.
        """
        loff, coff = offsets
        lfac, cfac = factors
        x__ = (col - coff) / cfac * 2**16
        y__ = (line - loff) / lfac * 2**16

        return x__, y__

    def get_area_extent(self, size, offsets, factors, platform_height):
        """Get the area extent of the file."""
        nlines, ncols = size
        h = platform_height

        # count starts at 1
        cols = 1 - 0.5
        lines = 1 - 0.5
        ll_x, ll_y = self.get_xy_from_linecol(lines, cols, offsets, factors)

        cols += ncols
        lines += nlines
        ur_x, ur_y = self.get_xy_from_linecol(lines, cols, offsets, factors)

        return (np.deg2rad(ll_x) * h, np.deg2rad(ll_y) * h,
                np.deg2rad(ur_x) * h, np.deg2rad(ur_y) * h)

    def get_area_def(self, dsid):
        """Get the area definition of the band."""
        cfac = np.int32(self.mda["cfac"])
        lfac = np.int32(self.mda["lfac"])
        coff = np.float32(self.mda["coff"])
        loff = np.float32(self.mda["loff"])

        a = self.mda["projection_parameters"]["a"]
        b = self.mda["projection_parameters"]["b"]
        h = self.mda["projection_parameters"]["h"]
        lon_0 = self.mda["projection_parameters"]["SSP_longitude"]
        nlines = int(self.mda["number_of_lines"])
        ncols = int(self.mda["number_of_columns"])

        area_extent = self.get_area_extent((nlines, ncols),
                                           (loff, coff),
                                           (lfac, cfac),
                                           h)

        proj_dict = {"a": float(a),
                     "b": float(b),
                     "lon_0": float(lon_0),
                     "h": float(h),
                     "proj": "geos",
                     "units": "m"}

        area = geometry.AreaDefinition(
            "some_area_name",
            "On-the-fly area",
            "geosmsg",
            proj_dict,
            ncols,
            nlines,
            area_extent)

        self.area = area
        return area

    def read_band(self, key, info):
        """Read the data."""
        output_dtype, output_shape = self._get_output_info()
        return da.from_delayed(_read_data(self.filename, self.mda),
                               shape=output_shape,
                               dtype=output_dtype)

    def _get_output_info(self):
        bpp = self.mda["number_of_bits_per_pixel"]
        if bpp in [10, 16]:
            output_dtype = np.uint16
        elif bpp == 8:
            output_dtype = np.uint8
        else:
            raise ValueError(f"Unexpected number of bits per pixel: {bpp}")
        output_shape = (self.mda["number_of_lines"], self.mda["number_of_columns"])
        return output_dtype, output_shape


@dask.delayed
def _read_data(filename, mda):
    return HRITSegment(filename, mda).read_data()


class HRITSegment:
    """An HRIT segment with data."""

    def __init__(self, filename, mda):
        """Set up the segment."""
        self.filename = filename
        self.mda = mda
        self.lines = mda["number_of_lines"]
        self.cols = mda["number_of_columns"]
        self.bpp = mda["number_of_bits_per_pixel"]
        self.compressed = mda["compression_flag_for_data"] == 1
        self.offset = mda["total_header_length"]

    def read_data(self):
        """Read the data."""
        data = self._read_data_from_file()
        if self.bpp == 10:
            data = dec10216(data)
        data = data.reshape((self.lines, self.cols))
        return data

    def _read_data_from_file(self):
        if self._is_file_like():
            return self._read_file_like()
        return self._read_data_from_disk()

    def _is_file_like(self):
        return not isinstance(self.filename, str)

    def _read_data_from_disk(self):
        # For reading the image data, unzip_context is faster than generic_open
        dtype, shape = self._get_input_info()
        with utils.unzip_context(self.filename) as fn:

            if self.compressed:
                return np.frombuffer(
                    decompress_file(fn),
                    offset=self.offset,
                    dtype=dtype,
                    count=np.prod(shape)
                )
            else:
                return np.fromfile(
                    fn,
                    offset=self.offset,
                    dtype=dtype,
                    count=np.prod(shape)
                )

    def _read_file_like(self):
        # filename is likely to be a file-like object, already in memory
        dtype, shape = self._get_input_info()
        with utils.generic_open(self.filename, mode="rb") as fp:
            decompressed_buffer = fp.read()
            if self.compressed:
                decompressed_buffer = decompress_buffer(decompressed_buffer)
            no_elements = np.prod(shape)
            return np.frombuffer(
                decompressed_buffer,
                dtype=np.dtype(dtype),
                count=no_elements.item(),
                offset=self.offset
            ).reshape(shape)

    def _get_input_info(self):
        total_bits = int(self.lines) * int(self.cols) * int(self.bpp)
        input_shape = int(np.ceil(total_bits / 8.))
        if self.bpp == 16:
            input_dtype = ">u2"
            input_shape //= 2
        elif self.bpp in [8, 10]:
            input_dtype = np.uint8
        else:
            raise ValueError(f"Unexpected number of bits per pixel: {self.bpp}")
        input_shape = (input_shape,)
        return input_dtype, input_shape



#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Grouping functionality for the readers."""

import datetime as dt
import logging
import warnings

import yaml

from .config import configs_for_reader
from .loading import load_reader

LOG = logging.getLogger(__name__)


def group_files(files_to_sort, reader=None, time_threshold=10,
                group_keys=None, reader_kwargs=None,
                missing="pass"):
    """Group series of files by file pattern information.

    By default this will group files by their filename ``start_time``
    assuming it exists in the pattern. By passing the individual
    dictionaries returned by this function to the Scene classes'
    ``filenames``, a series `Scene` objects can be easily created.

    Args:
        files_to_sort (Iterable): File paths to sort in to group
        reader (str or Collection[str]): Reader or readers whose file patterns
            should be used to sort files.  If not given, try all readers (slow,
            adding a list of readers is strongly recommended).
        time_threshold (int): Number of seconds used to consider time elements
            in a group as being equal. For example, if the 'start_time' item
            is used to group files then any time within `time_threshold`
            seconds of the first file's 'start_time' will be seen as occurring
            at the same time.
        group_keys (list or tuple): File pattern information to use to group
            files. Keys are sorted in order and only the first key is used when
            comparing datetime elements with `time_threshold` (see above). This
            means it is recommended that datetime values should only come from
            the first key in ``group_keys``. Otherwise, there is a good chance
            that files will not be grouped properly (datetimes being barely
            unequal). Defaults to a reader's ``group_keys`` configuration (set
            in YAML), otherwise ``('start_time',)``.  When passing multiple
            readers, passing group_keys is strongly recommended as the
            behaviour without doing so is undefined.
        reader_kwargs (dict): Additional keyword arguments to pass to reader
            creation.
        missing (str): Parameter to control the behavior in the scenario where
            multiple readers were passed, but at least one group does not have
            files associated with every reader.  Valid values are ``"pass"``
            (the default), ``"skip"``, and ``"raise"``.  If set to ``"pass"``,
            groups are passed as-is.  Some groups may have zero files for some
            readers.  If set to ``"skip"``, groups for which one or more
            readers have zero files are skipped (meaning that some files may
            not be associated to any group).  If set to ``"raise"``, raise a
            `FileNotFoundError` in case there are any groups for which one or
            more readers have no files associated.

    Returns:
        List of dictionaries mapping 'reader' to a list of filenames.
        Each of these dictionaries can be passed as ``filenames`` to
        a `Scene` object.

    """
    if reader is not None and not isinstance(reader, (list, tuple)):
        reader = [reader]

    reader_kwargs = reader_kwargs or {}

    reader_files = _assign_files_to_readers(
            files_to_sort, reader, reader_kwargs)

    if reader is None:
        reader = reader_files.keys()

    file_keys = _get_file_keys_for_reader_files(
            reader_files, group_keys=group_keys)

    file_groups = _get_sorted_file_groups(file_keys, time_threshold)

    groups = [{rn: file_groups[group_key].get(rn, []) for rn in reader} for group_key in file_groups]

    return list(_filter_groups(groups, missing=missing))


def _assign_files_to_readers(files_to_sort, reader_names,  # noqa: D417
                             reader_kwargs):
    """Assign files to readers.

    Given a list of file names (paths), match those to reader instances.

    Internal helper for group_files.

    Args:
        files_to_sort (Collection[str]): Files to assign to readers.
        reader_names (Collection[str]): Readers to consider
        reader_kwargs (Mapping):

    Returns:
        Mapping[str, Tuple[reader, Set[str]]]
        Mapping where the keys are reader names and the values are tuples of
        (reader_configs, filenames).
    """
    files_to_sort = set(files_to_sort)
    reader_dict = {}
    for reader_configs in configs_for_reader(reader_names):
        try:
            reader = load_reader(reader_configs, **reader_kwargs)
        except yaml.constructor.ConstructorError:
            LOG.exception(
                    f"ConstructorError loading {reader_configs!s}, "
                    "probably a missing dependency, skipping "
                    "corresponding reader (if you did not explicitly "
                    "specify the reader, Satpy tries all; performance "
                    "will improve if you pass readers explicitly).")
            continue
        reader_name = reader.info["name"]
        files_matching = set(reader.filter_selected_filenames(files_to_sort))
        files_to_sort -= files_matching
        if files_matching or reader_names is not None:
            reader_dict[reader_name] = (reader, files_matching)
    if files_to_sort:
        raise ValueError("No matching readers found for these files: " +
                         ", ".join(files_to_sort))
    return reader_dict


def _get_file_keys_for_reader_files(reader_files, group_keys=None):
    """From a mapping from _assign_files_to_readers, get file keys.

    Given a mapping where each key is a reader name and each value is a
    tuple of reader instance (typically FileYAMLReader) and a collection
    of files, return a mapping with the same keys, but where the values are
    lists of tuples of (keys, filename), where keys are extracted from the filenames
    according to group_keys and filenames are the names those keys were
    extracted from.

    Internal helper for group_files.

    Returns:
        Mapping[str, List[Tuple[Tuple, str]]], as described.
    """
    file_keys = {}
    for (reader_name, (reader_instance, files_to_sort)) in reader_files.items():
        if group_keys is None:
            group_keys = reader_instance.info.get("group_keys", ("start_time",))
        file_keys[reader_name] = []
        # make a copy because filename_items_for_filetype will modify inplace
        files_to_sort = set(files_to_sort)
        _walk_through_sorted_filetype_items(reader_instance, file_keys, files_to_sort, group_keys, reader_name)
    return file_keys


def _walk_through_sorted_filetype_items(reader_instance, file_keys, files_to_sort, group_keys, reader_name):
    for _, filetype_info in reader_instance.sorted_filetype_items():
        for f, file_info in reader_instance.filename_items_for_filetype(files_to_sort, filetype_info):
            _update_file_keys(file_keys, group_keys, file_info, f, reader_name)


def _update_file_keys(file_keys, group_keys, file_info, f, reader_name):
    group_key = tuple(file_info.get(k) for k in group_keys)
    if all(g is None for g in group_key):
        warnings.warn(
            f"Found matching file {f:s} for reader "
            f"{reader_name:s}, but none of group keys found. "
            "Group keys requested: " + ", ".join(group_keys),
            UserWarning,
            stacklevel=5
        )
    file_keys[reader_name].append((group_key, f))


def _get_sorted_file_groups(all_file_keys, time_threshold):  # noqa: D417
    """Get sorted file groups.

    Get a list of dictionaries, where each list item consists of a dictionary
    mapping a tuple of keys to a mapping of reader names to files.  The files
    listed in each list item are considered to be grouped within the same time.

    Args:
        all_file_keys (Iterable): as returned by _get_file_keys_for_reader_files
        time_threshold (numbers.Number): temporal threshold in seconds

    Returns:
        List[Mapping[Tuple, Mapping[str, List[str]]]], as described

    Internal helper for group_files.
    """
    # flatten to get an overall sorting; put the name in the middle in the
    # interest of sorting
    flat_keys = ((v[0], rn, v[1]) for (rn, vL) in all_file_keys.items() for v in vL)
    prev_key = None
    threshold = dt.timedelta(seconds=time_threshold)
    # file_groups is sorted, because dictionaries are sorted by insertion
    # order in Python 3.7+
    file_groups = {}
    for gk, rn, f in sorted(flat_keys):
        # use first element of key as time identifier (if datetime type)
        if prev_key is None:
            is_new_group = True
            prev_key = gk
        else:
            is_new_group = _get_group_status(gk, prev_key, threshold)

        # compare keys for those that are found for both the key and
        # this is a generator and is not computed until the if statement below
        # when we know that `prev_key` is not None
        vals_not_equal = (this_val != prev_val for this_val, prev_val in zip(gk[1:], prev_key[1:])
                          if this_val is not None and prev_val is not None)
        # if this is a new group based on the first element
        if is_new_group or any(vals_not_equal):
            file_groups[gk] = {rn: [f]}
            prev_key = gk
        else:
            _update_existing_group(file_groups, rn, prev_key, f)
    return file_groups


def _get_group_status(gk, prev_key, threshold):
    if isinstance(gk[0], dt.datetime):
        # datetimes within threshold difference are "the same time"
        return (gk[0] - prev_key[0]) > threshold
    return gk[0] != prev_key[0]


def _update_existing_group(file_groups, rn, prev_key, f):
    if rn not in file_groups[prev_key]:
        file_groups[prev_key][rn] = [f]
    else:
        file_groups[prev_key][rn].append(f)


def _filter_groups(groups, missing="pass"):
    """Filter multi-reader group-files behavior.

    Helper for `group_files`.  When `group_files` is called with multiple
    readers, make sure that the desired behaviour for missing files is
    enforced: if missing is ``"raise"``, raise an exception if at least one
    group has at least one reader without files; if it is ``"skip"``, remove
    those.  If it is ``"pass"``, do nothing.  Yields groups to be kept.

    Args:
        groups (List[Mapping[str, List[str]]]):
            groups as found by `group_files`.
        missing (str):
            String controlling behaviour, see documentation above.

    Yields:
        ``Mapping[str:, List[str]]``: groups to be retained
    """
    if missing == "pass":
        yield from groups
        return
    if missing not in ("raise", "skip"):
        raise ValueError("Invalid value for ``missing`` argument.  Expected "
                         f"'raise', 'skip', or 'pass', got {missing!r}")
    for (i, grp) in enumerate(groups):
        readers_without_files = _get_keys_with_empty_values(grp)
        if readers_without_files:
            _check_raise_missing(missing, i, readers_without_files)
        else:
            yield grp


def _check_raise_missing(missing, i, readers_without_files):
    if missing == "raise":
        raise FileNotFoundError(
            f"when grouping files, group at index {i:d} "
            "had no files for readers: " +
            ", ".join(readers_without_files))


def _get_keys_with_empty_values(grp):
    """Find mapping keys where values have length zero.

    Helper for `_filter_groups`, which is in turn a helper for `group_files`.
    Given a mapping key -> Collection[Any], return the keys where the length of the
    collection is zero.

    Args:
        grp (Mapping[Any, Collection[Any]]): dictionary to check

    Returns:
        set of keys
    """
    empty = set()
    for (k, v) in grp.items():
        if len(v) == 0:  # explicit check to ensure failure if not a collection
            empty.add(k)
    return empty


def find_files_and_readers(start_time=None, end_time=None, base_dir=None,
                           reader=None, sensor=None,
                           filter_parameters=None, reader_kwargs=None,
                           missing_ok=False, fs=None):
    """Find files matching the provided parameters.

    Use `start_time` and/or `end_time` to limit found filenames by the times
    in the filenames (not the internal file metadata). Files are matched if
    they fall anywhere within the range specified by these parameters.

    Searching is **NOT** recursive.

    Files may be either on-disk or on a remote file system.  By default,
    files are searched for locally.  Users can search on remote filesystems by
    passing an instance of an implementation of
    `fsspec.spec.AbstractFileSystem` (strictly speaking, any object of a class
    implementing a ``glob`` method works).

    If locating files on a local file system, the returned dictionary
    can be passed directly to the `Scene` object through the `filenames`
    keyword argument.  If it points to a remote file system, it is the
    responsibility of the user to download the files first (directly
    reading from cloud storage is not currently available in Satpy).

    The behaviour of time-based filtering depends on whether or not the filename
    contains information about the end time of the data or not:

      - if the end time is not present in the filename, the start time of the filename
        is used and has to fall between (inclusive) the requested start and end times
      - otherwise, the timespan of the filename has to overlap the requested timespan

    Example usage for querying a s3 filesystem using the s3fs module:

    >>> import s3fs, satpy.readers, datetime
    >>> satpy.readers.find_files_and_readers(
    ...     base_dir="s3://noaa-goes16/ABI-L1b-RadF/2019/321/14/",
    ...     fs=s3fs.S3FileSystem(anon=True),
    ...     reader="abi_l1b",
    ...     start_time=datetime.datetime(2019, 11, 17, 14, 40))
    {'abi_l1b': [...]}

    Args:
        start_time (datetime.datetime): Limit used files by starting time.
        end_time (datetime.datetime): Limit used files by ending time.
        base_dir (str): The directory to search for files containing the
                        data to load. Defaults to the current directory.
        reader (str or list): The name of the reader to use for loading the data or a list of names.
        sensor (str or list): Limit used files by provided sensors.
        filter_parameters (dict): Filename pattern metadata to filter on. `start_time` and `end_time` are
                                  automatically added to this dictionary. Shortcut for
                                  `reader_kwargs['filter_parameters']`.
        reader_kwargs (dict): Keyword arguments to pass to specific reader
                              instances to further configure file searching.
        missing_ok (bool): If False (default), raise ValueError if no files
                            are found.  If True, return empty dictionary if no
                            files are found.
        fs (:class:`fsspec.spec.AbstractFileSystem`): Optional, instance of implementation of
            :class:`fsspec.spec.AbstractFileSystem` (strictly speaking, any object of a class implementing
            ``.glob`` is enough).  Defaults to searching the local filesystem.

    Returns:
        dict: Dictionary mapping reader name string to list of filenames

    """
    reader_files = {}
    reader_kwargs = reader_kwargs or {}
    filter_parameters = filter_parameters or reader_kwargs.get("filter_parameters", {})
    sensor_supported = False

    _set_filter_times(filter_parameters, start_time, end_time)
    reader_kwargs["filter_parameters"] = filter_parameters

    for reader_configs in configs_for_reader(reader):
        (reader_instance, loadables, this_sensor_supported) = _get_loadables_for_reader_config(
                base_dir, reader, sensor, reader_configs, reader_kwargs, fs)
        sensor_supported = sensor_supported or this_sensor_supported
        _update_reader_files(reader_files, reader_instance, loadables)

    _check_sensor_status(sensor, sensor_supported)
    _check_reader_file_status(reader_files, missing_ok)

    return reader_files


def _set_filter_times(filter_parameters, start_time, end_time):
    if start_time or end_time:
        filter_parameters["start_time"] = start_time
        filter_parameters["end_time"] = end_time


def _update_reader_files(reader_files, reader_instance, loadables):
    if loadables:
        reader_files[reader_instance.name] = list(loadables)


def _check_sensor_status(sensor, sensor_supported):
    if sensor and not sensor_supported:
        raise ValueError("Sensor '{}' not supported by any readers".format(sensor))


def _check_reader_file_status(reader_files, missing_ok):
    if not (reader_files or missing_ok):
        raise ValueError("No supported files found")


def _get_loadables_for_reader_config(base_dir, reader, sensor, reader_configs,
                                     reader_kwargs, fs):
    """Get loadables for reader configs.

    Helper for find_files_and_readers.

    Args:
        base_dir (str): as for `find_files_and_readers`
        reader (str): as for `find_files_and_readers`
        sensor (str): as for `find_files_and_readers`
        reader_configs (dict): reader metadata such as returned by
            `configs_for_reader`.
        reader_kwargs (dict): Keyword arguments to be passed to reader.
        fs (fsspec.spec.AbstractFileSystem): as for `find_files_and_readers`
    """
    sensor_supported = False
    reader_instance = _get_reader_instance(reader, reader_configs, **reader_kwargs)
    if isinstance(reader_instance, tuple):
        return reader_instance

    if not reader_instance.supports_sensor(sensor):
        return (reader_instance, [], False)
    if sensor is not None:
        # sensor was specified and a reader supports it
        sensor_supported = True

    loadables = _get_loadables_from_reader(reader_instance, base_dir, fs)
    return (reader_instance, loadables, sensor_supported)


def _get_reader_instance(reader, reader_configs, **reader_kwargs):
    try:
        return load_reader(reader_configs, **reader_kwargs)
    except (KeyError, IOError, yaml.YAMLError) as err:
        LOG.info("Cannot use %s", str(reader_configs))
        LOG.debug(str(err))
        if _is_single_reader(reader):
            # if it is a single reader then give a more usable error
            raise
        return (None, [], False)


def _is_single_reader(reader):
    return reader and (isinstance(reader, str) or len(reader) == 1)


def _get_loadables_from_reader(reader_instance, base_dir, fs):
    loadables = reader_instance.select_files_from_directory(base_dir, fs)
    if loadables:
        loadables = list(
            reader_instance.filter_selected_filenames(loadables))
    return loadables



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Compositors performing arithmetic operations."""

from __future__ import annotations

import logging

from satpy.dataset import combine_metadata

from .core import CompositeBase

LOG = logging.getLogger(__name__)


class DifferenceCompositor(CompositeBase):
    """Make the difference of two data arrays."""

    def __call__(self, projectables, nonprojectables=None, **attrs):
        """Generate the composite."""
        if len(projectables) != 2:
            raise ValueError("Expected 2 datasets, got %d" % (len(projectables),))
        projectables = self.match_data_arrays(projectables)
        info = combine_metadata(*projectables)
        info["name"] = self.attrs["name"]
        info.update(self.attrs)  # attrs from YAML/__init__
        info.update(attrs)  # overwriting of DataID properties

        proj = projectables[0] - projectables[1]
        proj.attrs = info
        return proj


class RatioCompositor(CompositeBase):
    """Make the ratio of two data arrays."""

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate the composite."""
        if len(projectables) != 2:
            raise ValueError("Expected 2 datasets, got %d" % (len(projectables),))
        projectables = self.match_data_arrays(projectables)
        info = combine_metadata(*projectables)
        info.update(self.attrs)

        proj = projectables[0] / projectables[1]
        proj.attrs = info
        return proj


class SumCompositor(CompositeBase):
    """Make the sum of two data arrays."""

    def __call__(self, projectables, nonprojectables=None, **info):
        """Generate the composite."""
        if len(projectables) != 2:
            raise ValueError("Expected 2 datasets, got %d" % (len(projectables),))
        projectables = self.match_data_arrays(projectables)
        info = combine_metadata(*projectables)
        info["name"] = self.attrs["name"]

        proj = projectables[0] + projectables[1]
        proj.attrs = info
        return proj



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for blending datasets with the Multiscene object."""

import datetime as dt

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy import DataQuery, Scene
from satpy.multiscene.blend_funcs import stack, timeseries
from satpy.tests.multiscene_tests.test_utils import (
    DEFAULT_SHAPE,
    _create_test_area,
    _create_test_dataset,
    _create_test_int8_dataset,
)
from satpy.tests.utils import make_dataid

NUM_TEST_ROWS = 2
NUM_TEST_COLS = 3


def _get_expected_stack_select(scene1: Scene, scene2: Scene) -> xr.DataArray:
    expected = scene2["polar-ct"]
    expected[..., NUM_TEST_ROWS, :] = scene1["geo-ct"][..., NUM_TEST_ROWS, :]
    expected[..., :, NUM_TEST_COLS] = scene1["geo-ct"][..., :, NUM_TEST_COLS]
    expected[..., -1, :] = scene1["geo-ct"][..., -1, :]
    return expected.compute()


def _get_expected_stack_blend(scene1: Scene, scene2: Scene) -> xr.DataArray:
    expected = scene2["polar-ct"].copy().compute().astype(np.float64)
    expected[..., NUM_TEST_ROWS, :] = 5 / 3  # (1*2 + 3*1) / (2 + 1)
    expected[..., :, NUM_TEST_COLS] = 5 / 3
    expected[..., -1, :] = np.nan  # (1*0 + 0*1) / (0 + 1)
    # weight of 1 is masked to 0 because invalid overlay value:
    expected[..., -1, NUM_TEST_COLS] = 2 / 2  # (1*2 + 0*1) / (2 + 0)
    return expected


@pytest.fixture
def test_area():
    """Get area definition used by test DataArrays."""
    return _create_test_area()


@pytest.fixture(params=[np.int8, np.float32])
def data_type(request):
    """Get array data type of the DataArray being tested."""
    return request.param


@pytest.fixture(params=["", "L", "RGB", "RGBA"])
def image_mode(request):
    """Get image mode of the main DataArray being tested."""
    return request.param


@pytest.fixture
def cloud_type_data_array1(test_area, data_type, image_mode):
    """Get DataArray for cloud type in the first test Scene."""
    dsid1 = make_dataid(
        name="geo-ct",
        resolution=3000,
        modifiers=()
    )
    shape = DEFAULT_SHAPE if len(image_mode) == 0 else (len(image_mode),) + DEFAULT_SHAPE
    dims = ("y", "x") if len(image_mode) == 0 else ("bands", "y", "x")
    if data_type is np.int8:
        data_arr = _create_test_int8_dataset(name="geo-ct", shape=shape, area=test_area, values=1, dims=dims)
    else:
        data_arr = _create_test_dataset(name="geo-ct", shape=shape, area=test_area, values=1.0, dims=dims)

    data_arr.attrs["platform_name"] = "Meteosat-11"
    data_arr.attrs["sensor"] = {"seviri"}
    data_arr.attrs["units"] = "1"
    data_arr.attrs["long_name"] = "NWC GEO CT Cloud Type"
    data_arr.attrs["orbital_parameters"] = {
        "satellite_nominal_altitude": 35785863.0,
        "satellite_nominal_longitude": 0.0,
        "satellite_nominal_latitude": 0,
    }
    data_arr.attrs["start_time"] = dt.datetime(2023, 1, 16, 11, 9, 17)
    data_arr.attrs["end_time"] = dt.datetime(2023, 1, 16, 11, 12, 22)
    data_arr.attrs["_satpy_id"] = dsid1
    return data_arr


@pytest.fixture
def cloud_type_data_array2(test_area, data_type, image_mode):
    """Get DataArray for cloud type in the second test Scene."""
    dsid1 = make_dataid(
        name="polar-ct",
        resolution=1000,
        modifiers=()
    )
    shape = DEFAULT_SHAPE if len(image_mode) == 0 else (len(image_mode),) + DEFAULT_SHAPE
    dims = ("y", "x") if len(image_mode) == 0 else ("bands", "y", "x")
    if data_type is np.int8:
        data_arr = _create_test_int8_dataset(name="polar-ct", shape=shape, area=test_area, values=3, dims=dims)
        data_arr[..., -1, :] = data_arr.attrs["_FillValue"]
    else:
        data_arr = _create_test_dataset(name="polar-ct", shape=shape, area=test_area, values=3.0, dims=dims)
        data_arr[..., -1, :] = np.nan
    data_arr.attrs["platform_name"] = "NOAA-18"
    data_arr.attrs["sensor"] = {"avhrr-3"}
    data_arr.attrs["units"] = "1"
    data_arr.attrs["long_name"] = "SAFNWC PPS CT Cloud Type"
    data_arr.attrs["start_time"] = dt.datetime(2023, 1, 16, 11, 12, 57, 500000)
    data_arr.attrs["end_time"] = dt.datetime(2023, 1, 16, 11, 28, 1, 900000)
    data_arr.attrs["_satpy_id"] = dsid1
    return data_arr


@pytest.fixture
def scene1_with_weights(cloud_type_data_array1, test_area):
    """Create first test scene with a dataset of weights."""
    from satpy import Scene

    scene = Scene()
    scene[cloud_type_data_array1.attrs["_satpy_id"]] = cloud_type_data_array1

    wgt1 = _create_test_dataset(name="geo-ct-wgt", area=test_area, values=0)

    wgt1[NUM_TEST_ROWS, :] = 2
    wgt1[:, NUM_TEST_COLS] = 2

    dsid2 = make_dataid(
        name="geo-cma",
        resolution=3000,
        modifiers=()
    )
    scene[dsid2] = _create_test_int8_dataset(name="geo-cma", area=test_area, values=2)
    scene[dsid2].attrs["start_time"] = dt.datetime(2023, 1, 16, 11, 9, 17)
    scene[dsid2].attrs["end_time"] = dt.datetime(2023, 1, 16, 11, 12, 22)

    wgt2 = _create_test_dataset(name="geo-cma-wgt", area=test_area, values=0)

    return scene, [wgt1, wgt2]


@pytest.fixture
def scene2_with_weights(cloud_type_data_array2, test_area):
    """Create second test scene."""
    from satpy import Scene

    scene = Scene()
    scene[cloud_type_data_array2.attrs["_satpy_id"]] = cloud_type_data_array2

    wgt1 = _create_test_dataset(name="polar-ct-wgt", area=test_area, values=1)

    dsid2 = make_dataid(
        name="polar-cma",
        resolution=1000,
        modifiers=()
    )
    scene[dsid2] = _create_test_int8_dataset(name="polar-cma", area=test_area, values=4)
    scene[dsid2].attrs["start_time"] = dt.datetime(2023, 1, 16, 11, 12, 57, 500000)
    scene[dsid2].attrs["end_time"] = dt.datetime(2023, 1, 16, 11, 28, 1, 900000)

    wgt2 = _create_test_dataset(name="polar-cma-wgt", area=test_area, values=1)
    return scene, [wgt1, wgt2]


@pytest.fixture
def multi_scene_and_weights(scene1_with_weights, scene2_with_weights):
    """Create small multi-scene for testing."""
    from satpy import MultiScene
    scene1, weights1 = scene1_with_weights
    scene2, weights2 = scene2_with_weights

    return MultiScene([scene1, scene2]), [weights1, weights2]


@pytest.fixture
def groups():
    """Get group definitions for the MultiScene."""
    return {
        DataQuery(name="CloudType"): ["geo-ct", "polar-ct"],
        DataQuery(name="CloudMask"): ["geo-cma", "polar-cma"]
    }


class TestBlendFuncs:
    """Test individual functions used for blending."""

    def test_blend_two_scenes_using_stack(self, multi_scene_and_weights, groups,
                                          scene1_with_weights, scene2_with_weights):
        """Test blending two scenes by stacking them on top of each other using function 'stack'."""
        multi_scene, weights = multi_scene_and_weights
        scene1, weights1 = scene1_with_weights
        scene2, weights2 = scene2_with_weights

        multi_scene.group(groups)

        resampled = multi_scene
        stacked = resampled.blend(blend_function=stack)
        result = stacked["CloudType"].compute()

        expected = scene2["polar-ct"].copy()
        expected[..., -1, :] = scene1["geo-ct"][..., -1, :]

        xr.testing.assert_equal(result, expected.compute())
        _check_stacked_metadata(result, "CloudType")
        assert result.attrs["start_time"] == dt.datetime(2023, 1, 16, 11, 9, 17)
        assert result.attrs["end_time"] == dt.datetime(2023, 1, 16, 11, 28, 1, 900000)

    def test_blend_two_scenes_bad_blend_type(self, multi_scene_and_weights, groups):
        """Test exception is raised when bad 'blend_type' is used."""
        from functools import partial

        multi_scene, weights = multi_scene_and_weights

        simple_groups = {DataQuery(name="CloudType"): groups[DataQuery(name="CloudType")]}
        multi_scene.group(simple_groups)
        weights = [weights[0][0], weights[1][0]]
        stack_func = partial(stack, weights=weights, blend_type="i_dont_exist")
        with pytest.raises(ValueError, match="Unknown weighted blending type: .*.Expected one of: .*"):
            multi_scene.blend(blend_function=stack_func)

    @pytest.mark.parametrize(
        ("blend_func", "exp_result_func"),
        [
            ("select_with_weights", _get_expected_stack_select),
            ("blend_with_weights", _get_expected_stack_blend),
        ])
    def test_blend_two_scenes_using_stack_weighted(self, multi_scene_and_weights, groups,
                                                   scene1_with_weights, scene2_with_weights,
                                                   blend_func, exp_result_func):
        """Test stacking two scenes using weights.

        Here we test that the start and end times can be combined so that they
        describe the start and times of the entire data series. We also test
        the various types of weighted stacking functions (ex. select, blend).

        """
        from functools import partial

        multi_scene, weights = multi_scene_and_weights
        scene1, weights1 = scene1_with_weights
        scene2, weights2 = scene2_with_weights

        simple_groups = {DataQuery(name="CloudType"): groups[DataQuery(name="CloudType")]}
        multi_scene.group(simple_groups)

        weights = [weights[0][0], weights[1][0]]
        stack_func = partial(stack, weights=weights, blend_type=blend_func)
        weighted_blend = multi_scene.blend(blend_function=stack_func)

        expected = exp_result_func(scene1, scene2)
        result = weighted_blend["CloudType"].compute()
        # result has NaNs and xarray's xr.testing.assert_equal doesn't support NaN comparison
        np.testing.assert_allclose(result.data, expected.data)

        _check_stacked_metadata(result, "CloudType")
        assert result.attrs["start_time"] == dt.datetime(2023, 1, 16, 11, 9, 17)
        assert result.attrs["end_time"] == dt.datetime(2023, 1, 16, 11, 28, 1, 900000)

    @pytest.fixture
    def datasets_and_weights(self):
        """X-Array datasets with area definition plus weights for input to tests."""
        shape = (8, 12)
        area = AreaDefinition("test", "test", "test",
                              {"proj": "geos", "lon_0": -95.5, "h": 35786023.0},
                              shape[1], shape[0], [-200, -200, 200, 200])

        ds1 = xr.DataArray(da.ones(shape, chunks=-1), dims=("y", "x"),
                           attrs={"start_time": dt.datetime(2018, 1, 1, 0, 0, 0), "area": area})
        ds2 = xr.DataArray(da.ones(shape, chunks=-1) * 2, dims=("y", "x"),
                           attrs={"start_time": dt.datetime(2018, 1, 1, 1, 0, 0), "area": area})
        ds3 = xr.DataArray(da.ones(shape, chunks=-1) * 3, dims=("y", "x"),
                           attrs={"start_time": dt.datetime(2018, 1, 1, 1, 0, 0), "area": area})

        ds4 = xr.DataArray(da.zeros(shape, chunks=-1), dims=("y", "time"),
                           attrs={"start_time": dt.datetime(2018, 1, 1, 0, 0, 0), "area": area})
        ds5 = xr.DataArray(da.zeros(shape, chunks=-1), dims=("y", "time"),
                           attrs={"start_time": dt.datetime(2018, 1, 1, 1, 0, 0), "area": area})

        wgt1 = xr.DataArray(da.ones(shape, chunks=-1), dims=("y", "x"),
                            attrs={"start_time": dt.datetime(2018, 1, 1, 0, 0, 0), "area": area})
        wgt2 = xr.DataArray(da.zeros(shape, chunks=-1), dims=("y", "x"),
                            attrs={"start_time": dt.datetime(2018, 1, 1, 0, 0, 0), "area": area})
        wgt3 = xr.DataArray(da.zeros(shape, chunks=-1), dims=("y", "x"),
                            attrs={"start_time": dt.datetime(2018, 1, 1, 0, 0, 0), "area": area})

        datastruct = {"shape": shape,
                      "area": area,
                      "datasets": [ds1, ds2, ds3, ds4, ds5],
                      "weights": [wgt1, wgt2, wgt3]}
        return datastruct

    @pytest.mark.parametrize(("line", "column",),
                             [(2, 3), (4, 5)]
                             )
    def test_blend_function_stack_weighted(self, datasets_and_weights, line, column):
        """Test the 'stack_weighted' function."""
        from functools import partial

        from satpy.dataset import combine_metadata

        input_data = datasets_and_weights

        input_data["weights"][1][line, :] = 2
        input_data["weights"][2][:, column] = 2

        stack_with_weights = partial(stack, weights=input_data["weights"])
        blend_result = stack_with_weights(input_data["datasets"][0:3])

        ds1 = input_data["datasets"][0]
        ds2 = input_data["datasets"][1]
        ds3 = input_data["datasets"][2]
        expected = ds1.copy()
        expected[:, column] = ds3[:, column]
        expected[line, :] = ds2[line, :]
        expected.attrs = combine_metadata(*[x.attrs for x in input_data["datasets"][0:3]])

        xr.testing.assert_equal(blend_result.compute(), expected.compute())
        assert expected.attrs == blend_result.attrs

    def test_blend_function_stack(self, datasets_and_weights):
        """Test the 'stack' function."""
        input_data = datasets_and_weights

        ds1 = input_data["datasets"][0]
        ds2 = input_data["datasets"][1]

        res = stack([ds1, ds2])
        expected = ds2.copy()
        expected.attrs["start_time"] = ds1.attrs["start_time"]

        xr.testing.assert_equal(res.compute(), expected.compute())
        assert expected.attrs == res.attrs

    def test_timeseries(self, datasets_and_weights):
        """Test the 'timeseries' function."""
        input_data = datasets_and_weights

        ds1 = input_data["datasets"][0]
        ds2 = input_data["datasets"][1]
        ds4 = input_data["datasets"][2]
        ds4 = input_data["datasets"][3]
        ds5 = input_data["datasets"][4]

        res = timeseries([ds1, ds2])
        res2 = timeseries([ds4, ds5])
        assert isinstance(res, xr.DataArray)
        assert isinstance(res2, xr.DataArray)
        assert (2, ds1.shape[0], ds1.shape[1]) == res.shape
        assert (ds4.shape[0], ds4.shape[1]+ds5.shape[1]) == res2.shape


def _check_stacked_metadata(data_arr: xr.DataArray, exp_name: str) -> None:
    assert data_arr.attrs["units"] == "1"
    assert data_arr.attrs["name"] == exp_name
    if "_FillValue" in data_arr.attrs:
        assert data_arr.attrs["_FillValue"] == 255
        assert data_arr.attrs["valid_range"] == [1, 15]

    expected_area = _create_test_area()
    assert data_arr.attrs["area"] == expected_area

    # these metadata items don't match between all inputs
    assert "sensor" not in data_arr.attrs
    assert "platform_name" not in data_arr.attrs
    assert "long_name" not in data_arr.attrs


class TestTemporalRGB:
    """Test the temporal RGB blending method."""

    @pytest.fixture
    def nominal_data(self):
        """Return the input arrays for the nominal use case."""
        da1 = xr.DataArray([1, 0, 0], attrs={"start_time": dt.datetime(2023, 5, 22, 9, 0, 0)})
        da2 = xr.DataArray([0, 1, 0], attrs={"start_time": dt.datetime(2023, 5, 22, 10, 0, 0)})
        da3 = xr.DataArray([0, 0, 1], attrs={"start_time": dt.datetime(2023, 5, 22, 11, 0, 0)})

        return [da1, da2, da3]

    @pytest.fixture
    def expected_result(self):
        """Return the expected result arrays."""
        return [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

    @staticmethod
    def _assert_results(res, expected_start_time, expected_result):
        assert res.attrs["start_time"] == expected_start_time
        np.testing.assert_equal(res.data[0, :], expected_result[0])
        np.testing.assert_equal(res.data[1, :], expected_result[1])
        np.testing.assert_equal(res.data[2, :], expected_result[2])

    def test_nominal(self, nominal_data, expected_result):
        """Test that nominal usage with 3 datasets works."""
        from satpy.multiscene.blend_funcs import temporal_rgb

        res = temporal_rgb(nominal_data)

        self._assert_results(res, nominal_data[-1].attrs["start_time"], expected_result)

    def test_extra_datasets(self, nominal_data, expected_result):
        """Test that only the first three arrays affect the usage."""
        from satpy.multiscene.blend_funcs import temporal_rgb

        da4 = xr.DataArray([0, 0, 1], attrs={"start_time": dt.datetime(2023, 5, 22, 12, 0, 0)})

        res = temporal_rgb(nominal_data + [da4,])

        self._assert_results(res, nominal_data[-1].attrs["start_time"], expected_result)


@pytest.mark.parametrize("func", ["stack", "temporal_rgb", "timeseries"])
def test_blend_funcs_warns(func):
    """Test that there's a warning when importing from ABI base from the old location."""
    from satpy import multiscene

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        getattr(multiscene, func)



# Copyright (c) 2015-2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Composite classes for spectral adjustments."""

import logging

from satpy.composites.core import GenericCompositor
from satpy.dataset import combine_metadata

LOG = logging.getLogger(__name__)


class SpectralBlender(GenericCompositor):
    """Construct new channel by blending contributions from a set of channels.

    This class can be used to compute weighted average of different channels.
    Primarily it's used to correct the green band of AHI and FCI in order to
    allow for proper true color imagery.

    Below is an example used to generate a corrected green channel for AHI using a weighted average from
    three channels, with 63% contribution from the native green channel (B02), 29% from the red channel (B03)
    and 8% from the near-infrared channel (B04)::

      corrected_green:
        compositor: !!python/name:satpy.composites.spectral.SpectralBlender
        fractions: [0.63, 0.29, 0.08]
        prerequisites:
          - name: B02
            modifiers: [sunz_corrected, rayleigh_corrected]
          - name: B03
            modifiers: [sunz_corrected, rayleigh_corrected]
          - name: B04
            modifiers: [sunz_corrected, rayleigh_corrected]
        standard_name: toa_bidirectional_reflectance

    Other examples can be found in the``ahi.yaml`` composite file in the satpy distribution.
    """

    def __init__(self, *args, fractions=(), **kwargs):
        """Set default keyword argument values."""
        self.fractions = fractions
        super().__init__(*args, **kwargs)

    def __call__(self, projectables, optional_datasets=None, **attrs):
        """Blend channels in projectables using the weights in self.fractions."""
        if len(self.fractions) != len(projectables):
            raise ValueError("fractions and projectables must have the same length.")

        projectables = self.match_data_arrays(projectables)
        new_channel = sum(fraction * value for fraction, value in zip(self.fractions, projectables))
        new_channel.attrs = combine_metadata(*projectables)
        return super().__call__((new_channel,), **attrs)


class HybridGreen(SpectralBlender):
    """Corrector of the FCI or AHI green band.

    The green band in FCI and AHI (and other bands centered at 0.51 microns) deliberately
    misses the chlorophyll spectral reflectance local maximum at 0.55 microns
    in order to focus on aerosol and ash rather than on vegetation. This
    affects true colour RGBs, because vegetation looks brown rather than green
    and barren surface types typically gets a reddish hue.

    To correct for this the hybrid green approach proposed by Miller et al. (2016, :doi:`10.1175/BAMS-D-15-00154.2`)
    is used. The basic idea is to include some contribution also from the 0.86 micron
    channel, which is known for its sensitivity to vegetation. The formula used for this is::

      hybrid_green = (1 - F) * R(0.51) + F * R(0.86)

    where F is a constant value, that is set to 0.15 by default in Satpy.

    For example, the HybridGreen compositor can be used as follows to construct a hybrid green channel for
    AHI, with 15% contibution from the near-infrared 0.85 µm band (B04) and the remaining 85% from the native
    green 0.51 µm band (B02)::

      hybrid_green:
        compositor: !!python/name:satpy.composites.spectral.HybridGreen
        fraction: 0.15
        prerequisites:
          - name: B02
            modifiers: [sunz_corrected, rayleigh_corrected]
          - name: B04
            modifiers: [sunz_corrected, rayleigh_corrected]
        standard_name: toa_bidirectional_reflectance

    Other examples can be found in the ``ahi.yaml`` and ``ami.yaml`` composite
    files in the satpy distribution.
    """

    def __init__(self, *args, fraction=0.15, **kwargs):
        """Set default keyword argument values."""
        fractions = (1 - fraction, fraction)
        super().__init__(fractions=fractions, *args, **kwargs)


class NDVIHybridGreen(SpectralBlender):
    """Construct a NDVI-weighted hybrid green channel.

    This green band correction follows the same approach as the HybridGreen compositor, but with a dynamic blend
    factor `f` that depends on the pixel-level Normalized Differece Vegetation Index (NDVI). The higher the NDVI, the
    smaller the contribution from the nir channel will be, following a liner (default) or non-linear relationship
    between the two ranges `[ndvi_min, ndvi_max]` and `limits`.

    As an example, a new green channel using e.g. FCI data and the NDVIHybridGreen compositor can be defined like::

      ndvi_hybrid_green:
        compositor: !!python/name:satpy.composites.spectral.NDVIHybridGreen
        ndvi_min: 0.0
        ndvi_max: 1.0
        limits: [0.15, 0.05]
        strength: 1.0
        prerequisites:
          - name: vis_05
            modifiers: [sunz_corrected, rayleigh_corrected]
          - name: vis_06
            modifiers: [sunz_corrected, rayleigh_corrected]
          - name: vis_08
            modifiers: [sunz_corrected ]
        standard_name: toa_bidirectional_reflectance

    In this example, pixels with NDVI=0.0 will be a weighted average with 15% contribution from the
    near-infrared vis_08 channel and the remaining 85% from the native green vis_05 channel, whereas
    pixels with NDVI=1.0 will be a weighted average with 5% contribution from the near-infrared
    vis_08 channel and the remaining 95% from the native green vis_05 channel. For other values of
    NDVI a linear interpolation between these values will be performed.

    A strength larger or smaller than 1.0 will introduce a non-linear relationship between the two ranges
    `[ndvi_min, ndvi_max]` and `limits`. Hence, a higher strength (> 1.0) will result in a slower transition
    to higher/lower fractions at the NDVI extremes. Similarly, a lower strength (< 1.0) will result in a
    faster transition to higher/lower fractions at the NDVI extremes.
    """

    def __init__(self, *args, ndvi_min=0.0, ndvi_max=1.0, limits=(0.15, 0.05), strength=1.0, **kwargs):
        """Initialize class and set the NDVI limits, blending fraction limits and strength."""
        if strength <= 0.0:
            raise ValueError(f"Expected strength greater than 0.0, got {strength}.")

        self.ndvi_min = ndvi_min
        self.ndvi_max = ndvi_max
        self.limits = limits
        self.strength = strength
        super().__init__(*args, **kwargs)

    def __call__(self, projectables, optional_datasets=None, **attrs):
        """Construct the hybrid green channel weighted by NDVI."""
        LOG.info(f"Applying NDVI-weighted hybrid-green correction with limits [{self.limits[0]}, "
                 f"{self.limits[1]}] and strength {self.strength}.")

        projectables = self.match_data_arrays(projectables)

        ndvi = (projectables[2] - projectables[1]) / (projectables[2] + projectables[1])

        ndvi = ndvi.clip(self.ndvi_min, self.ndvi_max)

        # Introduce non-linearity to ndvi for non-linear scaling to NIR blend fraction
        if self.strength != 1.0:  # self._apply_strength() has no effect if strength = 1.0 -> no non-linear behaviour
            ndvi = self._apply_strength(ndvi)

        # Compute pixel-level NIR blend fractions from ndvi
        fraction = self._compute_blend_fraction(ndvi)

        # Prepare input as required by parent class (SpectralBlender)
        self.fractions = (1 - fraction, fraction)

        return super().__call__([projectables[0], projectables[2]], **attrs)

    def _apply_strength(self, ndvi):
        """Introduce non-linearity by applying strength factor.

        The method introduces non-linearity to the ndvi for a non-linear scaling from ndvi to blend fraction in
        `_compute_blend_fraction`. This can be used for a slower or faster transision to higher/lower fractions
        at the ndvi extremes. If strength equals 1.0, this operation has no effect on the ndvi.
        """
        ndvi = ndvi ** self.strength / (ndvi ** self.strength + (1 - ndvi) ** self.strength)

        return ndvi

    def _compute_blend_fraction(self, ndvi):
        """Compute pixel-level fraction of NIR signal to blend with native green signal.

        This method linearly scales the input ndvi values to pixel-level blend fractions within the range
        `[limits[0], limits[1]]` following this implementation <https://stats.stackexchange.com/a/281164>.
        """
        fraction = (ndvi - self.ndvi_min) / (self.ndvi_max - self.ndvi_min) * (self.limits[1] - self.limits[0]) \
            + self.limits[0]

        return fraction


# TODO: Turn this into a weighted RGB compositor
class NaturalEnh(GenericCompositor):
    """Enhanced version of natural color composite by Simon Proud.

    Args:
        ch16_w (float): weight for red channel (1.6 um). Default: 1.3
        ch08_w (float): weight for green channel (0.8 um). Default: 2.5
        ch06_w (float): weight for blue channel (0.6 um). Default: 2.2

    """

    def __init__(self, name, ch16_w=1.3, ch08_w=2.5, ch06_w=2.2,
                 *args, **kwargs):
        """Initialize the class."""
        self.ch06_w = ch06_w
        self.ch08_w = ch08_w
        self.ch16_w = ch16_w
        super(NaturalEnh, self).__init__(name, *args, **kwargs)

    def __call__(self, projectables, *args, **kwargs):
        """Generate the composite."""
        projectables = self.match_data_arrays(projectables)
        ch16 = projectables[0]
        ch08 = projectables[1]
        ch06 = projectables[2]

        ch1 = self.ch16_w * ch16 + self.ch08_w * ch08 + self.ch06_w * ch06
        ch1.attrs = ch16.attrs
        ch2 = ch08
        ch3 = ch06

        return super(NaturalEnh, self).__call__((ch1, ch2, ch3),
                                                *args, **kwargs)



"""Tests for image filters."""
import logging

import xarray as xr

from satpy.modifiers import ModifierBase

logger = logging.getLogger(__name__)


class Median(ModifierBase):
    """Apply a median filter to the band."""

    def __init__(self, median_filter_params, **kwargs):  # noqa: D417
        """Create the instance.

        Args:
            median_filter_params: The arguments to pass to dask-image's median_filter function. For example, {size: 3}
                                  makes give the median filter a kernel of size 3.

        """
        self.median_filter_params = median_filter_params
        super().__init__(**kwargs)

    def __call__(self, arrays, **info):
        """Get the median filtered band."""
        from dask_image.ndfilters import median_filter

        data = arrays[0]
        logger.debug(f"Apply median filtering with parameters {self.median_filter_params}.")
        res = xr.DataArray(median_filter(data.data, **self.median_filter_params),
                           dims=data.dims, attrs=data.attrs, coords=data.coords)
        self.apply_modifier_info(data, res)
        return res



# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the goes imager nc reader (EUMETSAT variant)."""

import unittest
from unittest import mock

import numpy as np
import xarray as xr

from satpy.readers.goes_imager_nc import is_vis_channel
from satpy.tests.utils import make_dataid


class GOESNCEUMFileHandlerRadianceTest(unittest.TestCase):
    """Tests for the radiances."""

    longMessage = True

    @mock.patch("satpy.readers.goes_imager_nc.xr")
    def setUp(self, xr_):
        """Set up the tests."""
        from satpy.readers.goes_imager_nc import CALIB_COEFS, GOESEUMNCFileHandler

        self.coefs = CALIB_COEFS["GOES-15"]
        self.all_coefs = CALIB_COEFS
        self.channels = sorted(self.coefs.keys())
        self.ir_channels = sorted([ch for ch in self.channels
                                   if not is_vis_channel(ch)])
        self.vis_channels = sorted([ch for ch in self.channels
                                    if is_vis_channel(ch)])

        # Mock file access to return a fake dataset.
        nrows = ncols = 300
        self.radiance = np.ones((1, nrows, ncols))  # IR channels
        self.lon = np.zeros((nrows, ncols))  # Dummy
        self.lat = np.repeat(np.linspace(-150, 150, nrows), ncols).reshape(
            nrows, ncols)  # Includes invalid values to be masked

        xr_.open_dataset.return_value = xr.Dataset(
            {"data": xr.DataArray(data=self.radiance, dims=("time", "yc", "xc")),
             "time": xr.DataArray(data=np.array([0], dtype="datetime64[ns]"),
                                  dims=("time",)),
             "bands": xr.DataArray(data=np.array([1]))},
            attrs={"Satellite Sensor": "G-15"})

        geo_data = xr.Dataset(
            {"lon": xr.DataArray(data=self.lon, dims=("yc", "xc")),
             "lat": xr.DataArray(data=self.lat, dims=("yc", "xc"))},
            attrs={"Satellite Sensor": "G-15"})

        # Instantiate reader using the mocked open_dataset() method
        self.reader = GOESEUMNCFileHandler(filename="dummy", filename_info={},
                                           filetype_info={}, geo_data=geo_data)

    def test_get_dataset_radiance(self):
        """Test getting the radiances."""
        for ch in self.channels:
            if not is_vis_channel(ch):
                radiance = self.reader.get_dataset(
                    key=make_dataid(name=ch, calibration="radiance"), info={})
                # ... this only compares the valid (unmasked) elements
                assert np.all(self.radiance == radiance.to_masked_array()), \
                    f"get_dataset() returns invalid radiance for channel {ch}"

    def test_calibrate(self):
        """Test whether the correct calibration methods are called."""
        for ch in self.channels:
            if not is_vis_channel(ch):
                calibs = {"brightness_temperature": "_calibrate_ir"}
                for calib, method in calibs.items():
                    with mock.patch.object(self.reader, method) as target_func:
                        self.reader.calibrate(data=self.reader.nc["data"],
                                              calibration=calib, channel=ch)
                        target_func.assert_called()

    def test_get_sector(self):
        """Test sector identification."""
        from satpy.readers.goes_imager_nc import (
            FULL_DISC,
            NORTH_HEMIS_EAST,
            NORTH_HEMIS_WEST,
            SOUTH_HEMIS_EAST,
            SOUTH_HEMIS_WEST,
            UNKNOWN_SECTOR,
        )
        shapes = {
            (2700, 5200): FULL_DISC,
            (1850, 3450): NORTH_HEMIS_EAST,
            (600, 3500): SOUTH_HEMIS_EAST,
            (1310, 3300): NORTH_HEMIS_WEST,
            (1099, 2800): SOUTH_HEMIS_WEST,
            (123, 456): UNKNOWN_SECTOR
        }
        for (nlines, ncols), sector_ref in shapes.items():
            for channel in ("00_7", "10_7"):
                sector = self.reader._get_sector(channel=channel, nlines=nlines,
                                                 ncols=ncols)
                assert sector == sector_ref, "Incorrect sector identification"


class GOESNCEUMFileHandlerReflectanceTest(unittest.TestCase):
    """Testing the reflectances."""

    longMessage = True

    @mock.patch("satpy.readers.goes_imager_nc.xr")
    def setUp(self, xr_):
        """Set up the tests."""
        from satpy.readers.goes_imager_nc import CALIB_COEFS, GOESEUMNCFileHandler

        self.coefs = CALIB_COEFS["GOES-15"]
        self.all_coefs = CALIB_COEFS
        self.channels = sorted(self.coefs.keys())
        self.ir_channels = sorted([ch for ch in self.channels
                                   if not is_vis_channel(ch)])
        self.vis_channels = sorted([ch for ch in self.channels
                                    if is_vis_channel(ch)])

        # Mock file access to return a fake dataset.
        nrows = ncols = 300
        self.reflectance = 50 * np.ones((1, nrows, ncols))  # Vis channel
        self.lon = np.zeros((nrows, ncols))  # Dummy
        self.lat = np.repeat(np.linspace(-150, 150, nrows), ncols).reshape(
            nrows, ncols)  # Includes invalid values to be masked

        xr_.open_dataset.return_value = xr.Dataset(
            {"data": xr.DataArray(data=self.reflectance, dims=("time", "yc", "xc")),
             "time": xr.DataArray(data=np.array([0], dtype="datetime64[ns]"),
                                  dims=("time",)),
             "bands": xr.DataArray(data=np.array([1]))},
            attrs={"Satellite Sensor": "G-15"})

        geo_data = xr.Dataset(
            {"lon": xr.DataArray(data=self.lon, dims=("yc", "xc")),
             "lat": xr.DataArray(data=self.lat, dims=("yc", "xc"))},
            attrs={"Satellite Sensor": "G-15"})

        # Instantiate reader using the mocked open_dataset() method
        self.reader = GOESEUMNCFileHandler(filename="dummy", filename_info={},
                                           filetype_info={}, geo_data=geo_data)

    def test_get_dataset_reflectance(self):
        """Test getting the reflectance."""
        for ch in self.channels:
            if is_vis_channel(ch):
                refl = self.reader.get_dataset(
                    key=make_dataid(name=ch, calibration="reflectance"), info={})
                # ... this only compares the valid (unmasked) elements
                assert np.all(self.reflectance == refl.to_masked_array()), \
                    f"get_dataset() returns invalid reflectance for channel {ch}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for GK-2B GOCI-II L2 products from NOSC.

For more information about the data, see: <https://www.nosc.go.kr/eng/boardContents/actionBoardContentsCons0028.do>
"""

import datetime as dt
import logging

import xarray as xr

from satpy.readers.core.netcdf import NetCDF4FileHandler

logger = logging.getLogger(__name__)

GROUPS_MAP = {
    "goci2_l2_ac": ["geophysical_data/RhoC", "geophysical_data/Rrs", "navigation_data"],
    "goci2_l2_iop": [
        "geophysical_data/a_total",
        "geophysical_data/bb_total",
        "navigation_data",
    ],
}


class GOCI2L2NCFileHandler(NetCDF4FileHandler):
    """File handler for GOCI-II L2 official data in netCDF format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super().__init__(filename, filename_info, filetype_info)

        self.attrs = self["/attrs"]
        self.nc = self._merge_navigation_data(filetype_info["file_type"])

        # Read metadata which are common to all datasets
        self.nlines = self.nc.sizes["number_of_lines"]
        self.ncols = self.nc.sizes["pixels_per_line"]
        self.coverage = filename_info["coverage"]

    def _merge_navigation_data(self, filetype):
        """Merge navigation data and geophysical data."""
        if filetype in GROUPS_MAP.keys():
            groups = GROUPS_MAP[filetype]
        else:
            groups = ["geophysical_data", "navigation_data"]
        return xr.merge([self[group] for group in groups])

    @property
    def start_time(self):
        """Start timestamp of the dataset."""
        date_str = self.attrs["observation_start_time"]
        return dt.datetime.strptime(date_str, "%Y%m%d_%H%M%S")

    @property
    def end_time(self):
        """End timestamp of the dataset."""
        date_str = self.attrs["observation_end_time"]
        return dt.datetime.strptime(date_str, "%Y%m%d_%H%M%S")

    def get_dataset(self, key, info):
        """Load a dataset."""
        var = info["file_key"]
        logger.debug("Reading in get_dataset %s.", var)
        variable = self.nc[var]

        variable = variable.rename({"number_of_lines": "y", "pixels_per_line": "x"})

        # Some products may miss lon/lat standard_name, use name as base name if it is not already present
        if variable.attrs.get("standard_name", None) is None:
            variable.attrs.update({"standard_name": variable.name})

        variable.attrs.update(key.to_dict())
        return variable



"""GMS-5 VISSR archive data format.

Reference: `VISSR Format Description`_

.. _VISSR Format Description:
    https://www.data.jma.go.jp/mscweb/en/operation/fig/VISSR_FORMAT_GMS-5.pdf
"""

import numpy as np

U1 = ">u1"
I2 = ">i2"
I4 = ">i4"
R4 = ">f4"
R8 = ">f8"

VIS_CHANNEL = "VIS"
IR_CHANNEL = "IR"
CHANNEL_TYPES = {
    "VIS": VIS_CHANNEL,
    "IR1": IR_CHANNEL,
    "IR2": IR_CHANNEL,
    "IR3": IR_CHANNEL,
    "WV": IR_CHANNEL,
}
ALT_CHANNEL_NAMES = {"VIS": "VIS", "IR1": "IR1", "IR2": "IR2", "IR3": "WV"}
BLOCK_SIZE_VIS = 13504
BLOCK_SIZE_IR = 3664

IMAGE_PARAM_ITEM_SIZE = 2688
TIME = [("date", I4), ("time", I4)]
CHANNELS = [("VIS", R4), ("IR1", R4), ("IR2", R4), ("WV", R4)]
VISIR_SOLAR = [("VIS", R4), ("IR", R4)]

CONTROL_BLOCK = np.dtype([("control_block_size", I2),
                          ("head_block_number_of_parameter_block", I2),
                          ("parameter_block_size", I2),
                          ("head_block_number_of_image_data", I2),
                          ("total_block_size_of_image_data", I2),
                          ("available_block_size_of_image_data", I2),
                          ("head_valid_line_number", I2),
                          ("final_valid_line_number", I2),
                          ("final_data_block_number", I2)])

MODE_BLOCK_FRAME_PARAMETERS = [("bit_length", I4),
                               ("number_of_lines", I4),
                               ("number_of_pixels", I4),
                               ("stepping_angle", R4),
                               ("sampling_angle", R4),
                               ("lcw_pixel_size", I4),
                               ("doc_pixel_size", I4),
                               ("reserved", I4)]

MODE_BLOCK = np.dtype([("satellite_number", I4),
                       ("satellite_name", "|S12"),
                       ("observation_time_ad", "|S16"),
                       ("observation_time_mjd", R8),
                       ("gms_operation_mode", I4),
                       ("dpc_operation_mode", I4),
                       ("vissr_observation_mode", I4),
                       ("scanner_selection", I4),
                       ("sensor_selection", I4),
                       ("sensor_mode", I4),
                       ("scan_frame_mode", I4),
                       ("scan_mode", I4),
                       ("upper_limit_of_scan_number", I4),
                       ("lower_limit_of_scan_number", I4),
                       ("equatorial_scan_line_number", I4),
                       ("spin_rate", R4),
                       ("vis_frame_parameters", MODE_BLOCK_FRAME_PARAMETERS),
                       ("ir_frame_parameters", MODE_BLOCK_FRAME_PARAMETERS),
                       ("satellite_height", R4),
                       ("earth_radius", R4),
                       ("ssp_longitude", R4),
                       ("reserved_1", I4, 9),
                       ("table_of_sensor_trouble", I4, 14),
                       ("reserved_2", I4, 36),
                       ("status_tables_of_data_relative_address_segment", I4, 60)])

COORDINATE_CONVERSION_PARAMETERS = np.dtype([
    ("data_segment", I4),
    ("data_validity", I4),
    ("data_generation_time", TIME),
    ("scheduled_observation_time", R8),
    ("stepping_angle_along_line", CHANNELS),
    ("sampling_angle_along_pixel", CHANNELS),
    ("central_line_number_of_vissr_frame", CHANNELS),
    ("central_pixel_number_of_vissr_frame", CHANNELS),
    ("pixel_difference_of_vissr_center_from_normal_position", CHANNELS),
    ("number_of_sensor_elements", CHANNELS),
    ("total_number_of_vissr_frame_lines", CHANNELS),
    ("total_number_of_vissr_frame_pixels", CHANNELS),
    ("vissr_misalignment", R4, (3,)),
    ("matrix_of_misalignment", R4, (3, 3)),
    ("parameters", [("judgement_of_observation_convergence_time", R4),
                    ("judgement_of_line_convergence", R4),
                    ("east_west_angle_of_sun_light_condense_prism", R4),
                    ("north_south_angle_of_sun_light_condense_prism", R4),
                    ("pi", R4),
                    ("pi_divided_by_180", R4),
                    ("180_divided_by_pi", R4),
                    ("equatorial_radius", R4),
                    ("oblateness_of_earth", R4),
                    ("eccentricity_of_earth_orbit", R4),
                    ("first_angle_of_vissr_observation_in_sdb", R4),
                    ("upper_limited_line_of_2nd_prism_for_vis_solar_observation", R4),
                    ("lower_limited_line_of_1st_prism_for_vis_solar_observation", R4),
                    ("upper_limited_line_of_3rd_prism_for_vis_solar_observation", R4),
                    ("lower_limited_line_of_2nd_prism_for_vis_solar_observation", R4)]),
    ("solar_stepping_angle_along_line", VISIR_SOLAR),
    ("solar_sampling_angle_along_pixel", VISIR_SOLAR),
    ("solar_center_line_of_vissr_frame", VISIR_SOLAR),
    ("solar_center_pixel_of_vissr_frame", VISIR_SOLAR),
    ("solar_pixel_difference_of_vissr_center_from_normal_position", VISIR_SOLAR),
    ("solar_number_of_sensor_elements", VISIR_SOLAR),
    ("solar_total_number_of_vissr_frame_lines", VISIR_SOLAR),
    ("solar_total_number_of_vissr_frame_pixels", VISIR_SOLAR),
    ("reserved_1", I4, 19),
    ("orbital_parameters", [("epoch_time", R8),
                            ("semi_major_axis", R8),
                            ("eccentricity", R8),
                            ("orbital_inclination", R8),
                            ("longitude_of_ascending_node", R8),
                            ("argument_of_perigee", R8),
                            ("mean_anomaly", R8),
                            ("longitude_of_ssp", R8),
                            ("latitude_of_ssp", R8)]),
    ("reserved_2", I4, 2),
    ("attitude_parameters", [("epoch_time", R8),
                             ("angle_between_z_axis_and_satellite_spin_axis_at_epoch_time", R8),
                             ("angle_change_rate_between_spin_axis_and_z_axis", R8),
                             ("angle_between_spin_axis_and_zy_axis", R8),
                             ("angle_change_rate_between_spin_axis_and_zt_axis", R8),
                             ("daily_mean_of_spin_rate", R8)]),
    ("reserved_3", I4, 529),
    ("correction_of_image_distortion", [("stepping_angle_along_line_of_ir1", R4),
                                        ("stepping_angle_along_line_of_ir2", R4),
                                        ("stepping_angle_along_line_of_wv", R4),
                                        ("stepping_angle_along_line_of_vis", R4),
                                        ("sampling_angle_along_pixel_of_ir1", R4),
                                        ("sampling_angle_along_pixel_of_ir2", R4),
                                        ("sampling_angle_along_pixel_of_wv", R4),
                                        ("sampling_angle_along_pixel_of_vis", R4),
                                        ("x_component_vissr_misalignment", R4),
                                        ("y_component_vissr_misalignment", R4)])
])

ATTITUDE_PREDICTION_DATA = np.dtype([("prediction_time_mjd", R8),
                                     ("prediction_time_utc", TIME),
                                     ("right_ascension_of_attitude", R8),
                                     ("declination_of_attitude", R8),
                                     ("sun_earth_angle", R8),
                                     ("spin_rate", R8),
                                     ("right_ascension_of_orbital_plane", R8),
                                     ("declination_of_orbital_plane", R8),
                                     ("reserved", R8),
                                     ("eclipse_flag", I4),
                                     ("spin_axis_flag", I4)])

ATTITUDE_PREDICTION = np.dtype([("data_segment", I4),
                                ("data_validity", I4),
                                ("data_generation_time", TIME),
                                ("start_time", R8),
                                ("end_time", R8),
                                ("prediction_interval_time", R8),
                                ("number_of_prediction", I4),
                                ("data_size", I4),
                                ("data", ATTITUDE_PREDICTION_DATA, (33,))])

ORBIT_PREDICTION_DATA = [("prediction_time_mjd", R8),
                         ("prediction_time_utc", TIME),
                         ("satellite_position_1950", R8, (3,)),
                         ("satellite_velocity_1950", R8, (3,)),
                         ("satellite_position_earth_fixed", R8, (3,)),
                         ("satellite_velocity_earth_fixed", R8, (3,)),
                         ("greenwich_sidereal_time", R8),
                         ("sat_sun_vector_1950", [("azimuth", R8),
                                                  ("elevation", R8)]),
                         ("sat_sun_vector_earth_fixed", [("azimuth", R8),
                                                         ("elevation", R8)]),
                         ("conversion_matrix", R8, (3, 3)),
                         ("moon_directional_vector", R8, (3,)),
                         ("satellite_position", [("ssp_longitude", R8),
                                                 ("ssp_latitude", R8),
                                                 ("satellite_height", R8)]),
                         ("eclipse_period_flag", I4),
                         ("reserved", I4)]

ORBIT_PREDICTION = np.dtype([("data_segment", I4),
                             ("data_validity", I4),
                             ("data_generation_time", TIME),
                             ("start_time", R8),
                             ("end_time", R8),
                             ("prediction_interval_time", R8),
                             ("number_of_prediction", I4),
                             ("data_size", I4),
                             ("data", ORBIT_PREDICTION_DATA, (9,))])

VIS_CALIBRATION_TABLE = np.dtype([
    ("channel_number", I4),
    ("data_validity", I4),
    ("updated_time", TIME),
    ("table_id", I4),
    ("brightness_albedo_conversion_table", R4, (64,)),
    ("vis_channel_staircase_brightness_data", R4, (6,)),
    ("coefficients_table_of_vis_staircase_regression_curve", R4, (10,)),
    ("brightness_table_for_calibration", [("universal_space_brightness", R4),
                                          ("solar_brightness", R4)]),
    ("calibration_uses_brightness_correspondence_voltage_chart", [("universal_space_voltage", R4),
                                                                  ("solar_voltage", R4)]),
    ("calibration_coefficients_of_radiation_observation", [("G", R4), ("V0", R4)]),
    ("reserved", I4, (9,))
 ])

VIS_CALIBRATION = np.dtype([("data_segment", I4),
                            ("data_validity", I4),
                            ("data_generation_time", TIME),
                            ("sensor_group", I4),
                            ("vis1_calibration_table", VIS_CALIBRATION_TABLE),
                            ("vis2_calibration_table", VIS_CALIBRATION_TABLE),
                            ("vis3_calibration_table", VIS_CALIBRATION_TABLE),
                            ("reserved", I4, (267,))])

TELEMETRY_DATA = np.dtype([
    ("shutter_temp", R4),
    ("redundant_mirror_temp", R4),
    ("primary_mirror_temp", R4),
    ("baffle_fw_temp", R4),
    ("baffle_af_temp", R4),
    ("15_volt_auxiliary_power_supply", R4),
    ("radiative_cooler_temp_1", R4),
    ("radiative_cooler_temp_2", R4),
    ("electronics_module_temp", R4),
    ("scan_mirror_temp", R4),
    ("shutter_cavity_temp", R4),
    ("primary_mirror_sealed_temp", R4),
    ("redundant_mirror_sealed_temp", R4),
    ("shutter_temp_2", R4),
    ("reserved", R4, (2,))
])

IR_CALIBRATION = np.dtype([
    ("data_segment", I4),
    ("data_validity", I4),
    ("updated_time", TIME),
    ("sensor_group", I4),
    ("table_id", I4),
    ("reserved_1", I4, (2,)),
    ("conversion_table_of_equivalent_black_body_radiation", R4, (256,)),
    ("conversion_table_of_equivalent_black_body_temperature", R4, (256,)),
    ("staircase_brightness_data", R4, (6,)),
    ("coefficients_table_of_staircase_regression_curve", R4, (10,)),
    ("brightness_data_for_calibration", [("brightness_of_space", R4),
                                         ("brightness_of_black_body_shutter", R4),
                                         ("reserved", R4)]),
    ("voltage_table_for_brightness_of_calibration", [("voltage_of_space", R4),
                                                     ("voltage_of_black_body_shutter", R4),
                                                     ("reserved", R4)]),
    ("calibration_coefficients_of_radiation_observation", [("G", R4), ("V0", R4)]),
    ("valid_shutter_temperature", R4),
    ("valid_shutter_radiation", R4),
    ("telemetry_data_table", TELEMETRY_DATA),
    ("flag_of_calid_shutter_temperature_calculation", I4),
    ("reserved_2", I4, (109,))
])

SIMPLE_COORDINATE_CONVERSION_TABLE = np.dtype([
    ("coordinate_conversion_table", I2, (1250,)),
    ("earth_equator_radius", R4),
    ("satellite_height", R4),
    ("stepping_angle", R4),
    ("sampling_angle", R4),
    ("ssp_latitude", R4),
    ("ssp_longitude", R4),
    ("ssp_line_number", R4),
    ("ssp_pixel_number", R4),
    ("pi", R4),
    ("line_correction_ir1_vis", R4),
    ("pixel_correction_ir1_vis", R4),
    ("line_correction_ir1_ir2", R4),
    ("pixel_correction_ir1_ir2", R4),
    ("line_correction_ir1_wv", R4),
    ("pixel_correction_ir1_wv", R4),
    ("reserved", R4, (32,)),
])

IMAGE_PARAMS = {
    "mode": {
        "dtype": MODE_BLOCK,
        "offset": {
            VIS_CHANNEL: 2 * BLOCK_SIZE_VIS,
            IR_CHANNEL: 2 * BLOCK_SIZE_IR
        }
    },
    "coordinate_conversion": {
        "dtype": COORDINATE_CONVERSION_PARAMETERS,
        "offset": {
            VIS_CHANNEL: 2 * BLOCK_SIZE_VIS + 2 * IMAGE_PARAM_ITEM_SIZE,
            IR_CHANNEL: 4 * BLOCK_SIZE_IR
        }
    },
    "attitude_prediction": {
        "dtype": ATTITUDE_PREDICTION,
        "offset": {
            VIS_CHANNEL: 2 * BLOCK_SIZE_VIS + 3 * IMAGE_PARAM_ITEM_SIZE,
            IR_CHANNEL: 5 * BLOCK_SIZE_IR
        },
        "preserve": "data"
    },
    "orbit_prediction_1": {
        "dtype": ORBIT_PREDICTION,
        "offset": {
            VIS_CHANNEL: 3 * BLOCK_SIZE_VIS,
            IR_CHANNEL: 6 * BLOCK_SIZE_IR
        },
        "preserve": "data"
    },
    "orbit_prediction_2": {
        "dtype": ORBIT_PREDICTION,
        "offset": {
            VIS_CHANNEL: 3 * BLOCK_SIZE_VIS + 1 * IMAGE_PARAM_ITEM_SIZE,
            IR_CHANNEL: 7 * BLOCK_SIZE_IR
        },
        "preserve": "data"
    },
    "vis_calibration": {
        "dtype": VIS_CALIBRATION,
        "offset": {
            VIS_CHANNEL: 3 * BLOCK_SIZE_VIS + 3 * IMAGE_PARAM_ITEM_SIZE,
            IR_CHANNEL: 9 * BLOCK_SIZE_IR
        },
        "preserve": "data"
    },
    "ir1_calibration": {
        "dtype": IR_CALIBRATION,
        "offset": {
            VIS_CHANNEL: 4 * BLOCK_SIZE_VIS,
            IR_CHANNEL: 10 * BLOCK_SIZE_IR
        },
    },
    "ir2_calibration": {
        "dtype": IR_CALIBRATION,
        "offset": {
            VIS_CHANNEL: 4 * BLOCK_SIZE_VIS + IMAGE_PARAM_ITEM_SIZE,
            IR_CHANNEL: 11 * BLOCK_SIZE_IR
        },
    },
    "wv_calibration": {
        "dtype": IR_CALIBRATION,
        "offset": {
            VIS_CHANNEL: 4 * BLOCK_SIZE_VIS + 2 * IMAGE_PARAM_ITEM_SIZE,
            IR_CHANNEL: 12 * BLOCK_SIZE_IR
        },
    },
    "simple_coordinate_conversion_table": {
        "dtype": SIMPLE_COORDINATE_CONVERSION_TABLE,
        "offset": {
            VIS_CHANNEL: 5 * BLOCK_SIZE_VIS + 2 * IMAGE_PARAM_ITEM_SIZE,
            IR_CHANNEL: 16 * BLOCK_SIZE_IR
        },
    }
}

LINE_CONTROL_WORD = np.dtype([
    ("data_id", U1, (4, )),
    ("line_number", I4),
    ("line_name", I4),
    ("error_line_flag", I4),
    ("error_message", I4),
    ("mode_error_flag", I4),
    ("scan_time", R8),
    ("beta_angle", R4),
    ("west_side_earth_edge", I4),
    ("east_side_earth_edge", I4),
    ("received_time_1", R8),  # Typo in format description (I*4)
    ("received_time_2", I4),
    ("reserved", U1, (8, ))
])

IMAGE_DATA_BLOCK_IR = np.dtype([("LCW", LINE_CONTROL_WORD),
                                ("DOC", U1, (256,)),  # Omitted
                                ("image_data", U1, 3344)])

IMAGE_DATA_BLOCK_VIS = np.dtype([("LCW", LINE_CONTROL_WORD),
                                 ("DOC", U1, (64,)),  # Omitted
                                 ("image_data", U1, (13376,))])

IMAGE_DATA = {
    VIS_CHANNEL: {
        "offset": 6 * BLOCK_SIZE_VIS,
        "dtype": IMAGE_DATA_BLOCK_VIS,
    },
    IR_CHANNEL: {
        "offset": 18 * BLOCK_SIZE_IR,
        "dtype": IMAGE_DATA_BLOCK_IR
    }
}



# -*- coding: utf-8 -*-
"""Create sphinx roles for referencing the DOI of a published paper.

Extension to add links to DOIs. With this extension you can use e.g.
:doi:`10.1016/S0022-2836(05)80360-2` in your documents. This will
create a link to a DOI resolver
(``https://doi.org/10.1016/S0022-2836(05)80360-2``).
The link caption will be the raw DOI.
You can also give an explicit caption, e.g.
:doi:`Basic local alignment search tool <10.1016/S0022-2836(05)80360-2>`.

:copyright: Copyright 2015  Jon Lund Steffensen. Based on extlinks by
    the Sphinx team.
:license: BSD.

"""

from docutils import nodes, utils
from sphinx.util.nodes import split_explicit_title


def doi_role(typ, rawtext, text, lineno, inliner, options=None, content=None):
    """Create a doi role."""
    if options is None:
        options = {}
    if content is None:
        content = []
    text = utils.unescape(text)
    has_explicit_title, title, part = split_explicit_title(text)
    full_url = "https://doi.org/" + part
    if not has_explicit_title:
        title = "DOI:" + part
    pnode = nodes.reference(title, title, internal=False, refuri=full_url)
    return [pnode], []


def arxiv_role(typ, rawtext, text, lineno, inliner, options=None, content=None):
    """Create an arxive role."""
    if options is None:
        options = {}
    if content is None:
        content = []
    text = utils.unescape(text)
    has_explicit_title, title, part = split_explicit_title(text)
    full_url = "https://arxiv.org/abs/" + part
    if not has_explicit_title:
        title = "arXiv:" + part
    pnode = nodes.reference(title, title, internal=False, refuri=full_url)
    return [pnode], []


def setup_link_role(app):
    """Set up the role link."""
    app.add_role("doi", doi_role, override=True)
    app.add_role("DOI", doi_role, override=True)
    app.add_role("arXiv", arxiv_role, override=True)
    app.add_role("arxiv", arxiv_role, override=True)


def setup(app):
    """Set up the app."""
    app.connect("builder-inited", setup_link_role)
    return {"version": "0.1", "parallel_read_safe": True}



# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for spectral correction compositors."""

import unittest
from unittest import mock

import dask
import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.composites.spectral import HybridGreen, NDVIHybridGreen, SpectralBlender
from satpy.tests.utils import CustomScheduler


class TestSpectralComposites:
    """Test composites for spectral channel corrections."""

    def setup_method(self):
        """Initialize channels."""
        rows = 5
        cols = 10
        self.c01 = xr.DataArray(da.zeros((rows, cols), chunks=25) + 0.20, dims=("y", "x"), attrs={"name": "C02"})
        self.c02 = xr.DataArray(da.zeros((rows, cols), chunks=25) + 0.30, dims=("y", "x"), attrs={"name": "C03"})
        self.c03 = xr.DataArray(da.zeros((rows, cols), chunks=25) + 0.40, dims=("y", "x"), attrs={"name": "C04"})

    def test_bad_lengths(self):
        """Test that error is raised if the amount of channels to blend does not match the number of weights."""
        comp = SpectralBlender("blended_channel", fractions=(0.3, 0.7), prerequisites=(0.51, 0.85),
                               standard_name="toa_bidirectional_reflectance")
        with pytest.raises(ValueError, match="fractions and projectables must have the same length."):
            comp((self.c01, self.c02, self.c03))

    def test_spectral_blender(self):
        """Test the base class for spectral blending of channels."""
        comp = SpectralBlender("blended_channel", fractions=(0.3, 0.4, 0.3), prerequisites=(0.51, 0.65, 0.85),
                               standard_name="toa_bidirectional_reflectance")
        res = comp((self.c01, self.c02, self.c03))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "blended_channel"
        assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"
        data = res.compute()
        np.testing.assert_allclose(data, 0.3)

    def test_hybrid_green(self):
        """Test hybrid green correction of the 'green' band."""
        comp = HybridGreen("hybrid_green", fraction=0.15, prerequisites=(0.51, 0.85),
                           standard_name="toa_bidirectional_reflectance")
        res = comp((self.c01, self.c03))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "hybrid_green"
        assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"
        data = res.compute()
        np.testing.assert_allclose(data, 0.23)


class TestNdviHybridGreenCompositor:
    """Test NDVI-weighted hybrid green correction of green band."""

    def setup_method(self):
        """Initialize channels."""
        coord_val = [1.0, 2.0]
        self.c01 = xr.DataArray(
            da.from_array(np.array([[0.25, 0.30], [0.20, 0.30]], dtype=np.float32), chunks=25),
            dims=("y", "x"), coords=[coord_val, coord_val], attrs={"name": "C02"})
        self.c02 = xr.DataArray(
            da.from_array(np.array([[0.25, 0.30], [0.25, 0.35]], dtype=np.float32), chunks=25),
            dims=("y", "x"), coords=[coord_val, coord_val], attrs={"name": "C03"})
        self.c03 = xr.DataArray(
            da.from_array(np.array([[0.35, 0.35], [0.28, 0.65]], dtype=np.float32), chunks=25),
            dims=("y", "x"), coords=[coord_val, coord_val], attrs={"name": "C04"})

    def test_ndvi_hybrid_green(self):
        """Test General functionality with linear scaling from ndvi to blend fraction."""
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = NDVIHybridGreen("ndvi_hybrid_green", limits=(0.15, 0.05), prerequisites=(0.51, 0.65, 0.85),
                                   standard_name="toa_bidirectional_reflectance")

            # Test General functionality with linear strength (=1.0)
            res = comp((self.c01, self.c02, self.c03))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "ndvi_hybrid_green"
        assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"
        data = res.values
        np.testing.assert_array_almost_equal(data, np.array([[0.2633, 0.3071], [0.2115, 0.3420]]), decimal=4)

    def test_ndvi_hybrid_green_dtype(self):
        """Test that the datatype is not altered by the compositor."""
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = NDVIHybridGreen("ndvi_hybrid_green", limits=(0.15, 0.05), prerequisites=(0.51, 0.65, 0.85),
                                   standard_name="toa_bidirectional_reflectance")
            res = comp((self.c01, self.c02, self.c03))
        assert res.data.dtype == np.float32

    def test_nonlinear_scaling(self):
        """Test non-linear scaling using `strength` term."""
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = NDVIHybridGreen("ndvi_hybrid_green", limits=(0.15, 0.05), strength=2.0,
                                   prerequisites=(0.51, 0.65, 0.85),
                                   standard_name="toa_bidirectional_reflectance")
            res = comp((self.c01, self.c02, self.c03))
        res_np = res.data.compute()
        assert res.dtype == res_np.dtype
        assert res.dtype == np.float32
        np.testing.assert_array_almost_equal(res.data, np.array([[0.2646, 0.3075], [0.2120, 0.3471]]), decimal=4)

    def test_invalid_strength(self):
        """Test using invalid `strength` term for non-linear scaling."""
        with pytest.raises(ValueError, match="Expected strength greater than 0.0, got 0.0."):
            _ = NDVIHybridGreen("ndvi_hybrid_green", strength=0.0, prerequisites=(0.51, 0.65, 0.85),
                                standard_name="toa_bidirectional_reflectance")

    def test_with_slightly_mismatching_coord_input(self):
        """Test the case where an input (typically the red band) has a slightly different coordinate.

        If match_data_arrays is called correctly, the coords will be aligned and the array will have the expected shape.

        """
        comp = NDVIHybridGreen("ndvi_hybrid_green", limits=(0.15, 0.05), prerequisites=(0.51, 0.65, 0.85),
                               standard_name="toa_bidirectional_reflectance")

        c02_bad_shape = self.c02.copy()
        c02_bad_shape.coords["y"] = [1.1, 2.]
        res = comp((self.c01, c02_bad_shape, self.c03))
        assert res.shape == (2, 2)


class TestNaturalEnhCompositor(unittest.TestCase):
    """Test NaturalEnh compositor."""

    def setUp(self):
        """Create channel data and set channel weights."""
        self.ch1 = xr.DataArray([1.0])
        self.ch2 = xr.DataArray([2.0])
        self.ch3 = xr.DataArray([3.0])
        self.ch16_w = 2.0
        self.ch08_w = 3.0
        self.ch06_w = 4.0

    @mock.patch("satpy.composites.spectral.NaturalEnh.__repr__")
    @mock.patch("satpy.composites.spectral.NaturalEnh.match_data_arrays")
    def test_natural_enh(self, match_data_arrays, repr_):
        """Test NaturalEnh compositor."""
        from satpy.composites.spectral import NaturalEnh
        repr_.return_value = ""
        projectables = [self.ch1, self.ch2, self.ch3]

        def temp_func(*args):
            return args[0]

        match_data_arrays.side_effect = temp_func
        comp = NaturalEnh("foo", ch16_w=self.ch16_w, ch08_w=self.ch08_w,
                          ch06_w=self.ch06_w)
        assert comp.ch16_w == self.ch16_w
        assert comp.ch08_w == self.ch08_w
        assert comp.ch06_w == self.ch06_w
        res = comp(projectables)
        assert mock.call(projectables) in match_data_arrays.mock_calls
        correct = (self.ch16_w * projectables[0] +
                   self.ch08_w * projectables[1] +
                   self.ch06_w * projectables[2])
        assert res[0] == correct
        assert res[1] == projectables[1]
        assert res[2] == projectables[2]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Classes related to the CREFL (corrected reflectance) modifier."""

import logging
import warnings

import numpy as np

from satpy.aux_download import DataDownloadMixin, retrieve
from satpy.modifiers import ModifierBase
from satpy.modifiers.angles import get_angles

LOG = logging.getLogger(__name__)


class ReflectanceCorrector(ModifierBase, DataDownloadMixin):
    """Corrected Reflectance (crefl) modifier.

    Uses a python rewrite of the C CREFL code written for VIIRS and MODIS.
    """

    def __init__(self, *args, dem_filename=None, dem_sds="averaged elevation",  # noqa: D417
                 url=None, known_hash=None, **kwargs):
        """Initialize the compositor with values from the user or from the configuration file.

        If `dem_filename` can't be found or opened then correction is done
        assuming TOA or sealevel options.

        Args:
            dem_filename (str): DEPRECATED
            url (str): URL or local path to the Digital Elevation Model (DEM)
                HDF4 file. If unset (None or empty string), then elevation
                is assumed to be 0 everywhere.
            known_hash (str): Optional SHA256 checksum to verify the download
                of ``url``.
            dem_sds (str): Name of the variable in the elevation file to load.

        """
        if dem_filename is not None:
            warnings.warn(
                "'dem_filename' for 'ReflectanceCorrector' is "
                "deprecated. Use 'url' instead.",
                DeprecationWarning,
                stacklevel=2
            )

        super(ReflectanceCorrector, self).__init__(*args, **kwargs)
        self.dem_sds = dem_sds
        self.url = url
        self.known_hash = known_hash
        self.dem_cache_key = self._get_registered_dem_cache_key()

    def _get_registered_dem_cache_key(self):
        if not self.url:
            return
        reg_files = self.register_data_files([{
            "url": self.url, "known_hash": self.known_hash}
        ])
        return reg_files[0]

    def __call__(self, datasets, optional_datasets, **info):
        """Create modified DataArray object by applying the crefl algorithm."""
        refl_data, angles = self._extract_angle_data_arrays(datasets, optional_datasets)
        results = self._call_crefl(refl_data, angles)
        info.update(refl_data.attrs)
        info["rayleigh_corrected"] = True
        results.attrs = info
        self.apply_modifier_info(refl_data, results)
        return results

    def _call_crefl(self, refl_data, angles):
        from satpy.modifiers._crefl_utils import run_crefl
        avg_elevation = self._get_average_elevation()
        results = run_crefl(refl_data,
                            *angles,
                            avg_elevation=avg_elevation,
                            )
        return results

    def _get_average_elevation(self):
        if self.dem_cache_key is None:
            return

        LOG.debug("Loading CREFL averaged elevation information from: %s",
                  self.dem_cache_key)
        local_filename = retrieve(self.dem_cache_key)
        avg_elevation = self._read_var_from_hdf4_file(local_filename, self.dem_sds).astype(np.float64)
        if isinstance(avg_elevation, np.ma.MaskedArray):
            avg_elevation = avg_elevation.filled(np.nan)
        return avg_elevation

    @staticmethod
    def _read_var_from_hdf4_file(local_filename, var_name):
        try:
            return ReflectanceCorrector._read_var_from_hdf4_file_pyhdf(local_filename, var_name)
        except (ImportError, OSError):
            return ReflectanceCorrector._read_var_from_hdf4_file_netcdf4(local_filename, var_name)

    @staticmethod
    def _read_var_from_hdf4_file_netcdf4(local_filename, var_name):
        from netCDF4 import Dataset as NCDataset

        # HDF4 file, NetCDF library needs to be compiled with HDF4 support
        nc = NCDataset(local_filename, "r")
        # average elevation is stored as a 16-bit signed integer but with
        # scale factor 1 and offset 0, convert it to float here
        return nc.variables[var_name][:]

    @staticmethod
    def _read_var_from_hdf4_file_pyhdf(local_filename, var_name):
        from pyhdf.SD import SD, SDC
        f = SD(local_filename, SDC.READ)
        var = f.select(var_name)
        data = var[:]
        fill = ReflectanceCorrector._read_fill_value_from_hdf4(var, data.dtype)
        return np.ma.MaskedArray(data, data == fill)

    @staticmethod
    def _read_fill_value_from_hdf4(var, dtype):
        from pyhdf.error import HDF4Error
        try:
            return var.getfillvalue()
        except HDF4Error:
            return np.iinfo(dtype).min

    def _extract_angle_data_arrays(self, datasets, optional_datasets):
        all_datasets = datasets + optional_datasets
        if len(all_datasets) == 1:
            vis = self.match_data_arrays(datasets)[0]
            return vis, get_angles(vis)
        if len(all_datasets) == 5:
            vis, *angles = self.match_data_arrays(
                datasets + optional_datasets)
            return vis, angles
        raise ValueError("Not sure how to handle provided dependencies. "
                         "Either all 4 angles must be provided or none of "
                         "of them.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2024.
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""A reader for Level 1C data produced by the MSI instrument aboard EarthCARE.

The MSI L1 files are described in the document "MSI L1 Product Definitions":
https://earthcarehandbook.earth.esa.int/documents/d/earthcare-data-handbook/msi-l1-product-definitions-volume-a-nominal-products

"""
import logging

import numpy as np

from satpy.readers.core.hdf5 import HDF5FileHandler
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


class MSIECL1CFileHandler(HDF5FileHandler):
    """File handler for MSI L1c H5 files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(MSIECL1CFileHandler, self).__init__(filename,
                                                  filename_info,
                                                  filetype_info)

    @property
    def end_time(self):
        """Get end time."""
        return self.filename_info["end_time"]

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["start_time"]

    def get_dataset(self, dataset_id, ds_info):
        """Load data variable and metadata and calibrate if needed."""
        file_key = ds_info.get("file_key", dataset_id["name"])
        data = self[file_key]

        # Band data is stored in a 3d array (Band x Along_Track x Across_Track).
        # This means we have to select a single 2d array for a given band,
        # and the correct index is given in the reader YAML.
        band_index = ds_info.get("band_index")
        if band_index is not None:
            data = data[band_index]

        # The dataset has incorrect units attribute (due to storing multiple types). Fix it here.
        data.attrs.update(ds_info)
        data.attrs.update({"units": ds_info.get("units")})

        if "fill_value" in ds_info:
            ds_fill = ds_info["fill_value"]
            data = data.where(data != ds_fill)

        # VIS/SWIR data can have radiance or reflectance calibration.
        if "calibration" in ds_info:
            cal_type = ds_info["calibration"].name
            data = self._calibrate(data, band_index, cal_type)

        # Rename dimensions, as some have incorrect names (notably the pixel value data).
        if "dim_1" in data.dims:
            data = data.rename({"dim_1": "y", "dim_2": "x"})

        return data

    def _calibrate(self, chan_data, band_index, cal_type):
        """Calibrate the data."""
        if cal_type == "reflectance":
            sol_irrad = self["ScienceData/solar_spectral_irradiance"]
            chan_data.data = chan_data.data * 100. * np.pi / sol_irrad[band_index].data.reshape((1, sol_irrad.shape[1]))
            return chan_data

        return chan_data
